/*
   BAREOS® - Backup Archiving REcovery Open Sourced

   Copyright (C) 2000-2011 Free Software Foundation Europe e.V.
   Copyright (C) 2011-2012 Planets Communications B.V.
   Copyright (C) 2013-2023 Bareos GmbH & Co. KG

   This program is Free Software; you can redistribute it and/or
   modify it under the terms of version three of the GNU Affero General Public
   License as published by the Free Software Foundation and included
   in the file LICENSE.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Affero General Public License for more details.

   You should have received a copy of the GNU Affero General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.
*/
// Kern Sibbald, March MM
/**
 * @file
 * Bareos File Daemon  backup.c  send file attributes and data to the Storage
 * daemon.
 */
#include "include/fcntl_def.h"
#include "include/bareos.h"
#include "include/filetypes.h"
#include "include/streams.h"
#include "filed/filed.h"
#include "filed/filed_globals.h"
#include "filed/accurate.h"
#include "filed/compression.h"
#include "filed/crypto.h"
#include "filed/heartbeat.h"
#include "filed/backup.h"
#include "filed/filed_jcr_impl.h"
#include "include/ch.h"
#include "findlib/attribs.h"
#include "findlib/hardlink.h"
#include "findlib/find_one.h"
#include "lib/attribs.h"
#include "lib/berrno.h"
#include "lib/bsock.h"
#include "lib/btimers.h"
#include "lib/parse_conf.h"
#include "lib/util.h"
#include "lib/serial.h"
#include "lib/compression.h"
#include "lib/channel.h"
#include "lib/thread_specific_data.h"

#include <thread>
#include <future>
#include <memory>

namespace filedaemon {

struct shared_buffer {
  std::size_t size{0};
  std::shared_ptr<char> data{nullptr};

  shared_buffer() = default;

  shared_buffer(std::size_t size) : size{size}
			   , data{new char[size], std::default_delete<char[]>{}}
  {}

  shared_buffer(std::shared_ptr<char>& lifetime,
		std::size_t size,
		char* data_) : size{size}
			     , data{lifetime, data_}
  {}
};

struct send_input {
  channel::out<shared_buffer> block;
  std::optional<channel::out<shared_buffer>> compressed;
  b_ctx* bctx;
  std::promise<void> barrier;
};
static void SendData(channel::out<shared_buffer> block,
		     std::optional<channel::out<shared_buffer>> compressed,
		     b_ctx* bctx,
		     std::promise<void> barrier);
void WaitForSend(JobControlRecord* jcr, channel::out<send_input> to_send) {
  SetJcrInThreadSpecificData(jcr);
  for (std::optional chan = to_send.get(); chan; chan = to_send.get()) {
    SendData(std::move(chan->block), std::move(chan->compressed),
	     chan->bctx, std::move(chan->barrier));
  }
}

struct digest_input {
  channel::out<shared_buffer> data;
  DIGEST* digest;
  DIGEST* signing_digest;
  std::promise<void> barrier;
};
static void DigestData(channel::out<shared_buffer> out,
		       DIGEST* digest, DIGEST* signing_digest,
		       std::promise<void> barrier);
void WaitForDigest(JobControlRecord* jcr, channel::out<digest_input> to_read) {
  SetJcrInThreadSpecificData(jcr);
  for (std::optional chan = to_read.get(); chan; chan = to_read.get()) {
    DigestData(std::move(chan->data), chan->digest, chan->signing_digest,
	       std::move(chan->barrier));
  }
}

struct compress_input {
  channel::out<shared_buffer> out;
  channel::in<shared_buffer> in;
  JobControlRecord* jcr;
  std::size_t max_compress_len;
  uint32_t algorithm;
  int      level;
};

struct compress_worker_input {
  compress_input args;
  std::mutex input{};
  std::mutex output{};
  std::condition_variable cv{};
  std::int64_t num_in{0};
  std::int64_t num_out{0};

  compress_worker_input(compress_input args) : args{std::move(args)}
  {}
};

static void Compress(CompressionContext& ctx,
		     channel::out<shared_buffer>& out,
		     std::mutex& input,
		     channel::in<shared_buffer>& in,
		     std::mutex& output,
		     std::condition_variable& cv,
		     std::int64_t& nin,
		     std::int64_t& nout,
		     JobControlRecord* jcr,
		     std::size_t max_compress_len,
		     uint32_t compression);
void WaitForCompress(JobControlRecord* jcr, channel::out<compress_input> to_read,
		     std::vector<channel::in<std::shared_ptr<compress_worker_input>>> pool) {
  SetJcrInThreadSpecificData(jcr);
  for (std::optional chan = to_read.get(); chan; chan = to_read.get()) {
    auto args = std::make_shared<compress_worker_input>(std::move(chan.value()));
    for (auto& worker : pool) {
      worker.put(args);
    }
  }
}
static void CompressionWorker(JobControlRecord* jcr,
			      channel::out<std::shared_ptr<compress_worker_input>> to_read,
			      std::unique_ptr<CompressionContext> ctx) {
  SetJcrInThreadSpecificData(jcr);
  for (;;) {
    if (std::optional chan = to_read.get()) {
      auto& work = chan.value();
      compress_input* args = &work->args;
      std::mutex& input = work->input;
      std::mutex& output = work->output;
      if (SetCompressionLevel(args->jcr,
			      args->algorithm,
			      args->level,
			      *ctx) == LevelChangeResult::CHANGE_ERROR) {
	continue;
      }
      Compress(*ctx,
	       args->out,
	       input,
	       args->in,
	       output,
	       work->cv,
	       work->num_in,
	       work->num_out,
	       args->jcr,
	       args->max_compress_len,
	       args->algorithm);
    } else {
      break;
    }
  }
}

struct SendContext {
  std::thread sender;
  std::thread digester;
  std::thread compressor;
  std::vector<std::thread> compress_workers;

  channel::in<send_input> to_send;
  channel::in<digest_input> to_digest;
  channel::in<compress_input> to_compress;


  SendContext(JobControlRecord* jcr,
	      std::size_t num_compress_workers) : compress_workers{} {
    if (num_compress_workers == 0) {
      // todo: warning message to jcr
      num_compress_workers = 1;
    }

    std::vector<channel::in<std::shared_ptr<compress_worker_input>>> compress_inputs{};
    for (std::size_t i = 0; i < num_compress_workers; ++i) {
      auto ctx = std::make_unique<CompressionContext>();
      auto [in, out] = channel::CreateBufferedChannel<std::shared_ptr<compress_worker_input>>(1);
      compress_inputs.emplace_back(std::move(in));
      AdjustCompressionBuffers(jcr, *ctx);
      compress_workers.emplace_back(CompressionWorker, jcr, std::move(out), std::move(ctx));
    }

    auto [sin, sout] = channel::CreateBufferedChannel<send_input>(1);
    auto [din, dout] = channel::CreateBufferedChannel<digest_input>(1);
    auto [cin, cout] = channel::CreateBufferedChannel<compress_input>(1);

    sender = std::thread{WaitForSend, jcr, std::move(sout)};
    digester = std::thread{WaitForDigest, jcr, std::move(dout)};
    compressor = std::thread{WaitForCompress, jcr, std::move(cout), std::move(compress_inputs)};

    to_send = std::move(sin);
    to_digest = std::move(din);
    to_compress = std::move(cin);
  }

  ~SendContext() {
    // we have to join the threads before we can destruct them;
    // we have to close the channels before we can join them
    to_send.close();
    to_digest.close();
    to_compress.close();
    sender.join();
    digester.join();
    compressor.join();
    for (auto& worker : compress_workers) {
      worker.join();
    }
  }
};

#ifdef HAVE_DARWIN_OS
const bool have_darwin_os = true;
#else
const bool have_darwin_os = false;
#endif

#if defined(HAVE_ACL)
const bool have_acl = true;
#else
const bool have_acl = false;
#endif

#if defined(HAVE_XATTR)
const bool have_xattr = true;
#else
const bool have_xattr = false;
#endif

#ifndef compressBound
#  define compressBound(sourceLen) \
    (sourceLen + (sourceLen >> 12) + (sourceLen >> 14) + (sourceLen >> 25) + 13)
#endif

/* Forward referenced functions */
static bool ShouldStripPaths(const FindFilesPacket* ff_pkt);
static bool do_strip(int count, const char* in, std::string& out);
static int send_data(JobControlRecord* jcr,
                     int stream,
                     FindFilesPacket* ff_pkt,
                     DIGEST* digest,
                     DIGEST* signature_digest);
bool EncodeAndSendAttributes(JobControlRecord* jcr,
                             FindFilesPacket* ff_pkt,
                             int& data_stream);
#if defined(WIN32_VSS)
static void CloseVssBackupSession(JobControlRecord* jcr);
#endif

std::pair<std::vector<channel::in<stated_file>>,
          std::vector<channel::out<stated_file>>>
CreateNecessaryChannels(findFILESET* fileset, std::size_t buffer_size)
{
  // Each include list in the fileset needs its own
  // channel.  These get created here.
  std::vector<channel::in<stated_file>> channel_ins;
  std::vector<channel::out<stated_file>> channel_outs;

  for (int i = 0; i < fileset->include_list.size(); ++i) {
    // channel ins/outs cannot be copied/ can only be moved.
    auto [in, out] = channel::CreateBufferedChannel<stated_file>(buffer_size);
    channel_ins.emplace_back(std::move(in));
    channel_outs.emplace_back(std::move(out));
  }

  return {std::move(channel_ins), std::move(channel_outs)};
}

/**
 * Find all the requested files and send them
 * to the Storage daemon.
 *
 * Note, we normally carry on a one-way
 * conversation from this point on with the SD, sfd_imply blasting
 * data to him.  To properly know what is going on, we
 * also run a "heartbeat" monitor which reads the socket and
 * reacts accordingly (at the moment it has nothing to do
 * except echo the heartbeat to the Director).
 */
bool BlastDataToStorageDaemon(JobControlRecord* jcr, crypto_cipher_t cipher)
{
  BareosSocket* sd;
  bool ok = true;

  sd = jcr->store_bsock;

  jcr->setJobStatusWithPriorityCheck(JS_Running);

  Dmsg1(300, "filed: opened data connection %d to stored\n", sd->fd_);
  ClientResource* client = nullptr;
  {
    ResLocker _{my_config};
    client = (ClientResource*)my_config->GetNextRes(R_CLIENT, NULL);
  }
  uint32_t buf_size;
  if (client) {
    buf_size = client->max_network_buffer_size;
  } else {
    buf_size = 0; /* use default */
  }
  if (!sd->SetBufferSize(buf_size, BNET_SETBUF_WRITE)) {
    jcr->setJobStatusWithPriorityCheck(JS_ErrorTerminated);
    Jmsg(jcr, M_FATAL, 0, _("Cannot set buffer size FD->SD.\n"));
    return false;
  }

  jcr->buf_size = sd->message_length;

  if (!AdjustCompressionBuffers(jcr, jcr->compress)) { return false; }

  if (!CryptoSessionStart(jcr, cipher)) { return false; }

  StartHeartbeatMonitor(jcr);

  if (have_acl) {
    jcr->fd_impl->acl_data = std::make_unique<AclData>();
    jcr->fd_impl->acl_data->u.build
        = (acl_build_data_t*)malloc(sizeof(acl_build_data_t));
    memset(jcr->fd_impl->acl_data->u.build, 0, sizeof(acl_build_data_t));
    jcr->fd_impl->acl_data->u.build->content = GetPoolMemory(PM_MESSAGE);
  }

  if (have_xattr) {
    jcr->fd_impl->xattr_data = std::make_unique<XattrData>();
    jcr->fd_impl->xattr_data->u.build
        = (xattr_build_data_t*)malloc(sizeof(xattr_build_data_t));
    memset(jcr->fd_impl->xattr_data->u.build, 0, sizeof(xattr_build_data_t));
    jcr->fd_impl->xattr_data->u.build->content = GetPoolMemory(PM_MESSAGE);
  }

  {
    // setup sending threads
    int32_t num_compression_threads = 4; // default
    if (client) {
      num_compression_threads = client->CompressionThreads;
    }
    SendContext data(jcr, num_compression_threads);
    jcr->fd_impl->send_ctx = &data;

#if 1 && !defined(SEND_SERIAL)

    // jcr->fd_impl->ff is only used by the sending thread
    // since it should just send the files it was given, we set
    // incremental & accurate to false so it does not double check them
    SetFindOptions((FindFilesPacket*)jcr->fd_impl->ff, false,
		   jcr->fd_impl->since_time);

    bool send_ok = true;
    try {
      std::optional results = ListFiles(jcr, jcr->fd_impl->ff->fileset,
					jcr->fd_impl->incremental,
					jcr->fd_impl->since_time,
					(jcr->accurate) ? std::make_optional(&AccurateCheckFile) : std::nullopt);

      if (results.has_value()) {
	for (std::optional found_files = results->out_chan().get();
	     found_files.has_value();
	     found_files = results->out_chan().get()) {
	  jcr->fd_impl->num_files_examined += found_files->count_skipped();
	  if (!SendFiles(jcr, jcr->fd_impl->ff, std::move(*found_files), SaveFile,
			 PluginSave)) {
	    send_ok = false;
	    break;
	  }
	}
      } else {
	send_ok = false;
      }

    } catch (const std::exception& ex)
      {
	send_ok = false;
	Dmsg1(500, "Uncaught exception! Aborting.\n Message: %s\n", ex.what());
      } catch (...)
      {
	send_ok = false;
	Dmsg1(500, "Uncaught unknown exception! Aborting.\n");
      }

    // join synchronizes threads, so its safe to read from list_ok now!
    if (!send_ok) {
      ok = false; /* error */
      jcr->setJobStatusWithPriorityCheck(JS_ErrorTerminated);
    }
#else
    SetFindOptions((FindFilesPacket*)jcr->fd_impl->ff, jcr->fd_impl->incremental,
		   jcr->fd_impl->since_time);

    // In accurate mode, we overload the find_one check function
    if (jcr->accurate) {
      SetFindChangedFunction((FindFilesPacket*)jcr->fd_impl->ff,
			     AccurateCheckFile);
    }
    // Subroutine SaveFile() is called for each file
    if (!FindFiles(jcr, (FindFilesPacket*)jcr->fd_impl->ff, SaveFile,
		   PluginSave)) {
      ok = false; /* error */
      jcr->setJobStatusWithPriorityCheck(JS_ErrorTerminated);
    }
#endif

    jcr->fd_impl->send_ctx = nullptr;
  }

  if (have_acl && jcr->fd_impl->acl_data->u.build->nr_errors > 0) {
    Jmsg(jcr, M_WARNING, 0,
         _("Encountered %ld acl errors while doing backup\n"),
         jcr->fd_impl->acl_data->u.build->nr_errors);
  }
  if (have_xattr && jcr->fd_impl->xattr_data->u.build->nr_errors > 0) {
    Jmsg(jcr, M_WARNING, 0,
         _("Encountered %ld xattr errors while doing backup\n"),
         jcr->fd_impl->xattr_data->u.build->nr_errors);
  }

#if defined(WIN32_VSS)
  CloseVssBackupSession(jcr);
#endif

  AccurateFinish(jcr); /* send deleted or base file list to SD */

  StopHeartbeatMonitor(jcr);

  sd->signal(BNET_EOD); /* end of sending data */

  if (have_acl && jcr->fd_impl->acl_data) {
    FreePoolMemory(jcr->fd_impl->acl_data->u.build->content);
    free(jcr->fd_impl->acl_data->u.build);
  }

  if (have_xattr && jcr->fd_impl->xattr_data) {
    FreePoolMemory(jcr->fd_impl->xattr_data->u.build->content);
    free(jcr->fd_impl->xattr_data->u.build);
  }

  if (jcr->fd_impl->big_buf) {
    free(jcr->fd_impl->big_buf);
    jcr->fd_impl->big_buf = NULL;
  }

  CleanupCompression(jcr);
  CryptoSessionEnd(jcr);

  Dmsg1(100, "end blast_data ok=%d\n", ok);
  return ok;
}

// Save OSX specific resource forks and finder info.
static inline bool SaveRsrcAndFinder(b_save_ctx& bsctx)
{
  char flags[FOPTS_BYTES];
  int rsrc_stream;
  BareosSocket* sd = bsctx.jcr->store_bsock;
  bool retval = false;

  if (bsctx.ff_pkt->hfsinfo.rsrclength > 0) {
    if (BopenRsrc(&bsctx.ff_pkt->bfd, bsctx.ff_pkt->fname, O_RDONLY | O_BINARY,
                  0)
        < 0) {
      bsctx.ff_pkt->ff_errno = errno;
      BErrNo be;
      Jmsg(bsctx.jcr, M_NOTSAVED, -1,
           _("     Cannot open resource fork for \"%s\": ERR=%s.\n"),
           bsctx.ff_pkt->fname, be.bstrerror());
      bsctx.jcr->JobErrors++;
      if (IsBopen(&bsctx.ff_pkt->bfd)) { bclose(&bsctx.ff_pkt->bfd); }
    } else {
      int status;

      memcpy(flags, bsctx.ff_pkt->flags, sizeof(flags));
      ClearBit(FO_COMPRESS, bsctx.ff_pkt->flags);
      ClearBit(FO_SPARSE, bsctx.ff_pkt->flags);
      ClearBit(FO_OFFSETS, bsctx.ff_pkt->flags);
      rsrc_stream = BitIsSet(FO_ENCRYPT, flags)
                        ? STREAM_ENCRYPTED_MACOS_FORK_DATA
                        : STREAM_MACOS_FORK_DATA;

      status = send_data(bsctx.jcr, rsrc_stream, bsctx.ff_pkt, bsctx.digest,
                         bsctx.signing_digest);

      memcpy(bsctx.ff_pkt->flags, flags, sizeof(flags));
      bclose(&bsctx.ff_pkt->bfd);
      if (!status) { goto bail_out; }
    }
  }

  Dmsg1(300, "Saving Finder Info for \"%s\"\n", bsctx.ff_pkt->fname);
  sd->fsend("%ld %d 0", bsctx.jcr->JobFiles, STREAM_HFSPLUS_ATTRIBUTES);
  Dmsg1(300, "filed>stored:header %s", sd->msg);
  PmMemcpy(sd->msg, bsctx.ff_pkt->hfsinfo.fndrinfo, 32);
  sd->message_length = 32;
  if (bsctx.digest) {
    CryptoDigestUpdate(bsctx.digest, (uint8_t*)sd->msg, sd->message_length);
  }
  if (bsctx.signing_digest) {
    CryptoDigestUpdate(bsctx.signing_digest, (uint8_t*)sd->msg,
                       sd->message_length);
  }
  sd->send();
  sd->signal(BNET_EOD);

  retval = true;

bail_out:
  return retval;
}

/**
 * Setup for digest handling. If this fails, the digest will be set to NULL
 * and not used. Note, the digest (file hash) can be any one of the four
 * algorithms below.
 *
 * The signing digest is a single algorithm depending on
 * whether or not we have SHA2.
 *   ****FIXME****  the signing algorithm should really be
 *   determined a different way!!!!!!  What happens if
 *   sha2 was available during backup but not restore?
 */
static inline bool SetupEncryptionDigests(b_save_ctx& bsctx)
{
  bool retval = false;
  // TODO landonf: Allow the user to specify the digest algorithm
#ifdef HAVE_SHA2
  crypto_digest_t signing_algorithm = CRYPTO_DIGEST_SHA256;
#else
  crypto_digest_t signing_algorithm = CRYPTO_DIGEST_SHA1;
#endif

  if (BitIsSet(FO_MD5, bsctx.ff_pkt->flags)) {
    bsctx.digest = crypto_digest_new(bsctx.jcr, CRYPTO_DIGEST_MD5);
    bsctx.digest_stream = STREAM_MD5_DIGEST;
  } else if (BitIsSet(FO_SHA1, bsctx.ff_pkt->flags)) {
    bsctx.digest = crypto_digest_new(bsctx.jcr, CRYPTO_DIGEST_SHA1);
    bsctx.digest_stream = STREAM_SHA1_DIGEST;
  } else if (BitIsSet(FO_SHA256, bsctx.ff_pkt->flags)) {
    bsctx.digest = crypto_digest_new(bsctx.jcr, CRYPTO_DIGEST_SHA256);
    bsctx.digest_stream = STREAM_SHA256_DIGEST;
  } else if (BitIsSet(FO_SHA512, bsctx.ff_pkt->flags)) {
    bsctx.digest = crypto_digest_new(bsctx.jcr, CRYPTO_DIGEST_SHA512);
    bsctx.digest_stream = STREAM_SHA512_DIGEST;
  } else if (BitIsSet(FO_XXH128, bsctx.ff_pkt->flags)) {
    bsctx.digest = crypto_digest_new(bsctx.jcr, CRYPTO_DIGEST_XXH128);
    bsctx.digest_stream = STREAM_XXH128_DIGEST;
  }

  // Did digest initialization fail?
  if (bsctx.digest_stream != STREAM_NONE && bsctx.digest == NULL) {
    Jmsg(bsctx.jcr, M_WARNING, 0, _("%s digest initialization failed\n"),
         stream_to_ascii(bsctx.digest_stream));
  }

  /* Set up signature digest handling. If this fails, the signature digest
   * will be set to NULL and not used. */
  /* TODO landonf: We should really only calculate the digest once, for
   * both verification and signing.
   */
  if (bsctx.jcr->fd_impl->crypto.pki_sign) {
    bsctx.signing_digest = crypto_digest_new(bsctx.jcr, signing_algorithm);

    // Full-stop if a failure occurred initializing the signature digest
    if (bsctx.signing_digest == NULL) {
      Jmsg(bsctx.jcr, M_NOTSAVED, 0,
           _("%s signature digest initialization failed\n"),
           stream_to_ascii(signing_algorithm));
      bsctx.jcr->JobErrors++;
      goto bail_out;
    }
  }

  // Enable encryption
  if (bsctx.jcr->fd_impl->crypto.pki_encrypt) {
    SetBit(FO_ENCRYPT, bsctx.ff_pkt->flags);
  }
  retval = true;

bail_out:
  return retval;
}

// Terminate the signing digest and send it to the Storage daemon
static inline bool TerminateSigningDigest(b_save_ctx& bsctx)
{
  uint32_t size = 0;
  bool retval = false;
  SIGNATURE* signature = NULL;
  BareosSocket* sd = bsctx.jcr->store_bsock;

  if ((signature = crypto_sign_new(bsctx.jcr)) == NULL) {
    Jmsg(bsctx.jcr, M_FATAL, 0,
         _("Failed to allocate memory for crypto signature.\n"));
    goto bail_out;
  }

  if (!CryptoSignAddSigner(signature, bsctx.signing_digest,
                           bsctx.jcr->fd_impl->crypto.pki_keypair)) {
    Jmsg(bsctx.jcr, M_FATAL, 0,
         _("An error occurred while signing the stream.\n"));
    goto bail_out;
  }

  // Get signature size
  if (!CryptoSignEncode(signature, NULL, &size)) {
    Jmsg(bsctx.jcr, M_FATAL, 0,
         _("An error occurred while signing the stream.\n"));
    goto bail_out;
  }

  // Grow the bsock buffer to fit our message if necessary
  if (SizeofPoolMemory(sd->msg) < (int32_t)size) {
    sd->msg = ReallocPoolMemory(sd->msg, size);
  }

  // Send our header
  sd->fsend("%ld %ld 0", bsctx.jcr->JobFiles, STREAM_SIGNED_DIGEST);
  Dmsg1(300, "filed>stored:header %s", sd->msg);

  // Encode signature data
  if (!CryptoSignEncode(signature, (uint8_t*)sd->msg, &size)) {
    Jmsg(bsctx.jcr, M_FATAL, 0,
         _("An error occurred while signing the stream.\n"));
    goto bail_out;
  }

  sd->message_length = size;
  sd->send();
  sd->signal(BNET_EOD); /* end of checksum */
  retval = true;

bail_out:
  if (signature) { CryptoSignFree(signature); }
  return retval;
}

// Terminate any digest and send it to Storage daemon
static inline bool TerminateDigest(b_save_ctx& bsctx)
{
  uint32_t size;
  bool retval = false;
  BareosSocket* sd = bsctx.jcr->store_bsock;

  sd->fsend("%ld %d 0", bsctx.jcr->JobFiles, bsctx.digest_stream);
  Dmsg1(300, "filed>stored:header %s", sd->msg);

  size = CRYPTO_DIGEST_MAX_SIZE;

  // Grow the bsock buffer to fit our message if necessary
  if (SizeofPoolMemory(sd->msg) < (int32_t)size) {
    sd->msg = ReallocPoolMemory(sd->msg, size);
  }

  if (!CryptoDigestFinalize(bsctx.digest, (uint8_t*)sd->msg, &size)) {
    Jmsg(bsctx.jcr, M_FATAL, 0,
         _("An error occurred finalizing signing the stream.\n"));
    goto bail_out;
  }

  // Keep the checksum if this file is a hardlink
  if (bsctx.ff_pkt->linked) {
    FfPktSetLinkDigest(bsctx.ff_pkt, bsctx.digest_stream, sd->msg, size);
  }

  sd->message_length = size;
  sd->send();
  sd->signal(BNET_EOD); /* end of checksum */
  retval = true;

bail_out:
  return retval;
}

static inline bool DoBackupAcl(JobControlRecord* jcr, FindFilesPacket* ff_pkt)
{
  bacl_exit_code retval;

  jcr->fd_impl->acl_data->filetype = ff_pkt->type;
  jcr->fd_impl->acl_data->last_fname = jcr->fd_impl->last_fname;

  if (jcr->IsPlugin()) {
    retval = PluginBuildAclStreams(jcr, jcr->fd_impl->acl_data.get(), ff_pkt);
  } else {
    retval = BuildAclStreams(jcr, jcr->fd_impl->acl_data.get(), ff_pkt);
  }

  switch (retval) {
    case bacl_exit_fatal:
      return false;
    case bacl_exit_error:
      Jmsg(jcr, M_ERROR, 0, "%s", jcr->errmsg);
      jcr->fd_impl->acl_data->u.build->nr_errors++;
      break;
    case bacl_exit_ok:
      break;
  }

  return true;
}

static inline bool DoBackupXattr(JobControlRecord* jcr, FindFilesPacket* ff_pkt)
{
  BxattrExitCode retval;

  jcr->fd_impl->xattr_data->last_fname = jcr->fd_impl->last_fname;

  if (jcr->IsPlugin()) {
    retval
        = PluginBuildXattrStreams(jcr, jcr->fd_impl->xattr_data.get(), ff_pkt);
  } else {
    retval = BuildXattrStreams(jcr, jcr->fd_impl->xattr_data.get(), ff_pkt);
  }

  switch (retval) {
    case BxattrExitCode::kErrorFatal:
      return false;
    case BxattrExitCode::kWarning:
      Jmsg(jcr, M_WARNING, 0, "%s", jcr->errmsg);
      break;
    case BxattrExitCode::kError:
      Jmsg(jcr, M_ERROR, 0, "%s", jcr->errmsg);
      jcr->fd_impl->xattr_data->u.build->nr_errors++;
      break;
    case BxattrExitCode::kSuccess:
      break;
  }

  return true;
}

static bool TerminateSaveFile(b_save_ctx& bsctx)
{
  FindFilesPacket* ff_pkt = bsctx.ff_pkt;
  JobControlRecord* jcr = bsctx.jcr;
  BareosSocket* sd = jcr->store_bsock;
  if (have_darwin_os) {
    // Regular files can have resource forks and Finder Info
    if (ff_pkt->type != FT_LNKSAVED
        && (S_ISREG(ff_pkt->statp.st_mode)
            && BitIsSet(FO_HFSPLUS, ff_pkt->flags))) {
      if (!SaveRsrcAndFinder(bsctx)) { return false; }
    }
  }

  // Save ACLs when requested and available for anything not being a symlink.
  if (have_acl) {
    if (BitIsSet(FO_ACL, ff_pkt->flags) && ff_pkt->type != FT_LNK) {
      if (!DoBackupAcl(jcr, ff_pkt)) { return false; }
    }
  }

  // Save Extended Attributes when requested and available for all files.
  if (have_xattr) {
    if (BitIsSet(FO_XATTR, ff_pkt->flags)) {
      if (!DoBackupXattr(jcr, ff_pkt)) { return false; }
    }
  }

  // Terminate the signing digest and send it to the Storage daemon
  if (bsctx.signing_digest) {
    if (!TerminateSigningDigest(bsctx)) { return false; }
  }

  // Terminate any digest and send it to Storage daemon
  if (bsctx.digest) {
    if (!TerminateDigest(bsctx)) { return false; }
  }

  // Check if original file has a digest, and send it
  if (ff_pkt->type == FT_LNKSAVED && ff_pkt->digest) {
    Dmsg2(300, "Link %s digest %d\n", ff_pkt->fname, ff_pkt->digest_len);
    sd->fsend("%ld %d 0", jcr->JobFiles, ff_pkt->digest_stream);

    sd->msg = CheckPoolMemorySize(sd->msg, ff_pkt->digest_len);
    memcpy(sd->msg, ff_pkt->digest, ff_pkt->digest_len);
    sd->message_length = ff_pkt->digest_len;
    sd->send();

    sd->signal(BNET_EOD); /* end of hardlink record */
  }

  return true;
}

/**
 * Called here by find() for each file included.
 * This is a callback. The original is FindFiles() above.
 *
 * Send the file and its data to the Storage daemon.
 *
 * Returns: 1 if OK
 *          0 if error
 *         -1 to ignore file/directory (not used here)
 */
int SaveFile(JobControlRecord* jcr, FindFilesPacket* ff_pkt, bool)
{
  bool do_read = false;
  bool plugin_started = false;
  bool do_plugin_set = false;
  int status, data_stream;
  int rtnstat = 0;
  b_save_ctx bsctx;
  bool has_file_data = false;
  save_pkt sp; /* use by option plugin */
  BareosSocket* sd = jcr->store_bsock;

  if (jcr->IsJobCanceled() || jcr->IsIncomplete()) { return 0; }

  jcr->fd_impl->num_files_examined++; /* bump total file count */

  switch (ff_pkt->type) {
    case FT_LNKSAVED: /* Hard linked, file already saved */
      Dmsg2(130, "FT_LNKSAVED hard link: %s => %s\n", ff_pkt->fname,
            ff_pkt->link);
      break;
    case FT_REGE:
      Dmsg1(130, "FT_REGE saving: %s\n", ff_pkt->fname);
      has_file_data = true;
      break;
    case FT_REG:
      Dmsg1(130, "FT_REG saving: %s\n", ff_pkt->fname);
      has_file_data = true;
      break;
    case FT_LNK:
      Dmsg2(130, "FT_LNK saving: %s -> %s\n", ff_pkt->fname, ff_pkt->link);
      break;
    case FT_RESTORE_FIRST:
      Dmsg1(100, "FT_RESTORE_FIRST saving: %s\n", ff_pkt->fname);
      break;
    case FT_PLUGIN_CONFIG:
      Dmsg1(100, "FT_PLUGIN_CONFIG saving: %s\n", ff_pkt->fname);
      break;
    case FT_DIRBEGIN:
      jcr->fd_impl->num_files_examined--; /* correct file count */
      return 1;                           /* not used */
    case FT_NORECURSE:
	    Jmsg(jcr, M_INFO, 1,
		 _("     Recursion turned off. Will not descend from %s into %s\n"),
		 ff_pkt->top_fname, ff_pkt->fname);
      ff_pkt->type = FT_DIREND; /* Backup only the directory entry */
      break;
    case FT_NOFSCHG:
      /* Suppress message for /dev filesystems */
      if (!IsInFileset(ff_pkt)) {
        Jmsg(jcr, M_INFO, 1,
             _("     %s is a different filesystem. Will not descend from %s "
               "into it.\n"),
             ff_pkt->fname, ff_pkt->top_fname);
      }
      ff_pkt->type = FT_DIREND; /* Backup only the directory entry */
      break;
    case FT_INVALIDFS:
      Jmsg(jcr, M_INFO, 1,
	   _("     Disallowed filesystem. Will not descend from %s into %s\n"),
	   ff_pkt->top_fname, ff_pkt->fname);
      ff_pkt->type = FT_DIREND; /* Backup only the directory entry */
      break;
    case FT_INVALIDDT:
      Jmsg(jcr, M_INFO, 1,
	   _("     Disallowed drive type. Will not descend into %s\n"),
	   ff_pkt->fname);
      break;
    case FT_REPARSE:
    case FT_JUNCTION:
    case FT_DIREND:
      Dmsg1(130, "FT_DIREND: %s\n", ff_pkt->link);
      break;
    case FT_SPEC:
      Dmsg1(130, "FT_SPEC saving: %s\n", ff_pkt->fname);
      if (S_ISSOCK(ff_pkt->statp.st_mode)) {
        Jmsg(jcr, M_SKIPPED, 1, _("     Socket file skipped: %s\n"),
             ff_pkt->fname);
        return 1;
      }
      break;
    case FT_RAW:
      Dmsg1(130, "FT_RAW saving: %s\n", ff_pkt->fname);
      has_file_data = true;
      break;
    case FT_FIFO:
      Dmsg1(130, "FT_FIFO saving: %s\n", ff_pkt->fname);
      break;
    case FT_NOACCESS: {
      BErrNo be;
      Jmsg(jcr, M_NOTSAVED, 0, _("     Could not access \"%s\": ERR=%s\n"),
           ff_pkt->fname, be.bstrerror(ff_pkt->ff_errno));
      jcr->JobErrors++;
      return 1;
    }
    case FT_NOFOLLOW: {
      BErrNo be;
      Jmsg(jcr, M_NOTSAVED, 0, _("     Could not follow link \"%s\": ERR=%s\n"),
           ff_pkt->fname, be.bstrerror(ff_pkt->ff_errno));
      jcr->JobErrors++;
      return 1;
    }
    case FT_NOSTAT: {
      BErrNo be;
      Jmsg(jcr, M_NOTSAVED, 0, _("     Could not stat \"%s\": ERR=%s\n"),
           ff_pkt->fname, be.bstrerror(ff_pkt->ff_errno));
      jcr->JobErrors++;
      return 1;
    }
    case FT_DIRNOCHG:
    case FT_NOCHG:
      Jmsg(jcr, M_SKIPPED, 1, _("     Unchanged file skipped: %s\n"),
           ff_pkt->fname);
      return 1;
    case FT_ISARCH:
      Jmsg(jcr, M_NOTSAVED, 0, _("     Archive file not saved: %s\n"),
           ff_pkt->fname);
      return 1;
    case FT_NOOPEN: {
      BErrNo be;
      Jmsg(jcr, M_NOTSAVED, 0,
           _("     Could not open directory \"%s\": ERR=%s\n"), ff_pkt->fname,
           be.bstrerror(ff_pkt->ff_errno));
      jcr->JobErrors++;
      return 1;
    }
    case FT_DELETED:
      Dmsg1(130, "FT_DELETED: %s\n", ff_pkt->fname);
      break;
    default:
      Jmsg(jcr, M_NOTSAVED, 0, _("     Unknown file type %d; not saved: %s\n"),
           ff_pkt->type, ff_pkt->fname);
      jcr->JobErrors++;
      return 1;
  }

  Dmsg1(130, "filed: sending %s to stored\n", ff_pkt->fname);

  // Setup backup signing context.
  memset(&bsctx, 0, sizeof(b_save_ctx));
  bsctx.digest_stream = STREAM_NONE;
  bsctx.jcr = jcr;
  bsctx.ff_pkt = ff_pkt;

  // Digests and encryption are only useful if there's file data
  if (has_file_data) {
    if (!SetupEncryptionDigests(bsctx)) { goto good_rtn; }
  }

  // Initialize the file descriptor we use for data and other streams.
  binit(&ff_pkt->bfd);
  if (BitIsSet(FO_PORTABLE, ff_pkt->flags)) {
    SetPortableBackup(&ff_pkt->bfd); /* disable Win32 BackupRead() */
  }

  // Option and cmd plugin are not compatible together
  if (ff_pkt->cmd_plugin) {
    do_plugin_set = true;
  } else if (ff_pkt->opt_plugin) {
    // Ask the option plugin what to do with this file
    switch (PluginOptionHandleFile(jcr, ff_pkt, &sp)) {
      case bRC_OK:
        Dmsg2(10, "Option plugin %s will be used to backup %s\n",
              ff_pkt->plugin, ff_pkt->fname);
        jcr->opt_plugin = true;
        jcr->fd_impl->plugin_sp = &sp;
        PluginUpdateFfPkt(ff_pkt, &sp);
        do_plugin_set = true;
        break;
      case bRC_Skip:
        Dmsg2(10, "Option plugin %s decided to skip %s\n", ff_pkt->plugin,
              ff_pkt->fname);
        goto good_rtn;
      case bRC_Core:
        Dmsg2(10, "Option plugin %s decided to let bareos handle %s\n",
              ff_pkt->plugin, ff_pkt->fname);
        break;
      default:
        goto bail_out;
    }
  }

  if (do_plugin_set) {
    // Tell bfile that it needs to call plugin
    if (!SetCmdPlugin(&ff_pkt->bfd, jcr)) { goto bail_out; }
    SendPluginName(jcr, sd, true); /* signal start of plugin data */
    plugin_started = true;
  }

  // Send attributes -- must be done after binit()
  if (!EncodeAndSendAttributes(jcr, ff_pkt, data_stream)) { goto bail_out; }

  // Meta data only for restore object
  if (IS_FT_OBJECT(ff_pkt->type)) { goto good_rtn; }

  // Meta data only for deleted files
  if (ff_pkt->type == FT_DELETED) { goto good_rtn; }

  // Set up the encryption context and send the session data to the SD
  if (has_file_data && jcr->fd_impl->crypto.pki_encrypt) {
    if (!CryptoSessionSend(jcr, sd)) { goto bail_out; }
  }

  /* For a command plugin use the setting from the plugins savepkt no_read field
   * which is saved in the ff_pkt->no_read variable. do_read is the inverted
   * value of this variable as no_read == TRUE means do_read == FALSE */
  if (ff_pkt->cmd_plugin) {
    do_read = !ff_pkt->no_read;
  } else {
    /* Open any file with data that we intend to save, then save it.
     *
     * Note, if is_win32_backup, we must open the Directory so that
     * the BackupRead will save its permissions and ownership streams. */
    if (ff_pkt->type != FT_LNKSAVED && S_ISREG(ff_pkt->statp.st_mode)) {
#ifdef HAVE_WIN32
      do_read = !IsPortableBackup(&ff_pkt->bfd) || ff_pkt->statp.st_size > 0;
#else
      do_read = ff_pkt->statp.st_size > 0;
#endif
    } else if (ff_pkt->type == FT_RAW || ff_pkt->type == FT_FIFO
               || ff_pkt->type == FT_REPARSE || ff_pkt->type == FT_JUNCTION
               || (!IsPortableBackup(&ff_pkt->bfd)
                   && ff_pkt->type == FT_DIREND)) {
      do_read = true;
    }
  }

  Dmsg2(150, "type=%d do_read=%d\n", ff_pkt->type, do_read);
  if (do_read) {
    btimer_t* tid;
    int noatime;

    if (ff_pkt->type == FT_FIFO) {
      tid = start_thread_timer(jcr, pthread_self(), 60);
    } else {
      tid = NULL;
    }

    noatime = BitIsSet(FO_NOATIME, ff_pkt->flags) ? O_NOATIME : 0;
    ff_pkt->bfd.reparse_point
        = (ff_pkt->type == FT_REPARSE || ff_pkt->type == FT_JUNCTION);

    if (bopen(&ff_pkt->bfd, ff_pkt->fname, O_RDONLY | O_BINARY | noatime, 0,
	      ff_pkt->statp.st_rdev) < 0) {
      ff_pkt->ff_errno = errno;
      BErrNo be;
      Jmsg(jcr, M_NOTSAVED, 0, _("     Cannot open \"%s\": ERR=%s.\n"),
           ff_pkt->fname, be.bstrerror());
      jcr->JobErrors++;
      if (tid) {
        StopThreadTimer(tid);
        tid = NULL;
      }
      goto good_rtn;
    }

    if (tid) {
      StopThreadTimer(tid);
      tid = NULL;
    }

    status = send_data(jcr, data_stream, ff_pkt, bsctx.digest,
                       bsctx.signing_digest);

    if (BitIsSet(FO_CHKCHANGES, ff_pkt->flags)) { HasFileChanged(jcr, ff_pkt); }

    bclose(&ff_pkt->bfd);

    if (!status) { goto bail_out; }
  }

  if (!TerminateSaveFile(bsctx)) {
    goto bail_out;
  }

good_rtn:
  rtnstat = jcr->IsJobCanceled() ? 0 : 1; /* good return if not canceled */

bail_out:
  if (jcr->IsIncomplete() || jcr->IsJobCanceled()) { rtnstat = 0; }
  if (plugin_started) {
    SendPluginName(jcr, sd, false); /* signal end of plugin data */
  }
  if (ff_pkt->opt_plugin) {
    jcr->fd_impl->plugin_sp = NULL; /* sp is local to this function */
    jcr->opt_plugin = false;
  }
  if (bsctx.digest) { CryptoDigestFree(bsctx.digest); }
  if (bsctx.signing_digest) { CryptoDigestFree(bsctx.signing_digest); }

  return rtnstat;
}



/**
 * Handle the data just read and send it to the SD after doing any
 * postprocessing needed.
 */
static inline bool SendDataToSd(b_ctx* bctx, std::optional<shared_buffer> precompressed = std::nullopt)
{
  BareosSocket* sd = bctx->jcr->store_bsock;
  bool need_more_data;

  // Check for sparse blocks
  if (BitIsSet(FO_SPARSE, bctx->ff_pkt->flags)) {
    bool allZeros;
    ser_declare;

    allZeros = false;
    if ((sd->message_length == bctx->rsize
         && (bctx->fileAddr + sd->message_length
             < (uint64_t)bctx->ff_pkt->statp.st_size))
        || ((bctx->ff_pkt->type == FT_RAW || bctx->ff_pkt->type == FT_FIFO)
            && ((uint64_t)bctx->ff_pkt->statp.st_size == 0))) {
      allZeros = IsBufZero(bctx->rbuf, bctx->rsize);
    }

    if (!allZeros) {
      // Put file address as first data in buffer
      SerBegin(bctx->wbuf, OFFSET_FADDR_SIZE);
      ser_uint64(bctx->fileAddr); /* store fileAddr in begin of buffer */
    }

    bctx->fileAddr += sd->message_length; /* update file address */

    // Skip block of all zeros
    if (allZeros) { return true; }
  } else if (BitIsSet(FO_OFFSETS, bctx->ff_pkt->flags)) {
    ser_declare;
    SerBegin(bctx->wbuf, OFFSET_FADDR_SIZE);
    ser_uint64(bctx->ff_pkt->bfd.offset); /* store offset in begin of buffer */
  }

  bctx->jcr->ReadBytes += sd->message_length; /* count bytes read */

  // Uncompressed cipher input length
  bctx->cipher_input_len = sd->message_length;

  // Update checksum if requested
  if (bctx->digest) {
    CryptoDigestUpdate(bctx->digest, (uint8_t*)bctx->rbuf, sd->message_length);
  }

  // Update signing digest if requested
  if (bctx->signing_digest) {
    CryptoDigestUpdate(bctx->signing_digest, (uint8_t*)bctx->rbuf,
                       sd->message_length);
  }

  // Compress the data.
  if (BitIsSet(FO_COMPRESS, bctx->ff_pkt->flags)) {
    if (precompressed) {
      memcpy(bctx->cbuf, precompressed->data.get(), precompressed->size);
      bctx->compress_len = precompressed->size;
    } else if (!CompressData(bctx->jcr, bctx->jcr->compress,
			     bctx->ff_pkt->Compress_algo, bctx->rbuf,
			     bctx->jcr->store_bsock->message_length, bctx->cbuf,
			     bctx->max_compress_len, &bctx->compress_len)) {
      return false;
    }

    // See if we need to generate a compression header.
    if (bctx->chead) {
      ser_declare;

      // Complete header
      SerBegin(bctx->chead, sizeof(comp_stream_header));
      ser_uint32(bctx->ch.magic);
      ser_uint32(bctx->compress_len);
      ser_uint16(bctx->ch.level);
      ser_uint16(bctx->ch.version);
      SerEnd(bctx->chead, sizeof(comp_stream_header));

      bctx->compress_len += sizeof(comp_stream_header); /* add size of header */
    }

    bctx->jcr->store_bsock->message_length
        = bctx->compress_len; /* set compressed length */
    bctx->cipher_input_len = bctx->compress_len;
  }

  // Encrypt the data.
  need_more_data = false;
  if (BitIsSet(FO_ENCRYPT, bctx->ff_pkt->flags)
      && !EncryptData(bctx, &need_more_data)) {
    if (need_more_data) { return true; }
    return false;
  }

  // Send the buffer to the Storage daemon
  if (BitIsSet(FO_SPARSE, bctx->ff_pkt->flags)
      || BitIsSet(FO_OFFSETS, bctx->ff_pkt->flags)) {
    sd->message_length += OFFSET_FADDR_SIZE; /* include fileAddr in size */
  }
  sd->msg = bctx->wbuf; /* set correct write buffer */

  if (!sd->send()) {
    if (!bctx->jcr->IsJobCanceled()) {
      Jmsg1(bctx->jcr, M_FATAL, 0, _("Network send error to SD. ERR=%s\n"),
            sd->bstrerror());
    }
    return false;
  }

  Dmsg1(130, "Send data to SD len=%d\n", sd->message_length);
  bctx->jcr->JobBytes += sd->message_length; /* count bytes saved possibly
                                                compressed/encrypted */
  sd->msg = bctx->msgsave;                   /* restore read buffer */

  return true;
}

#ifdef HAVE_WIN32
// Callback method for ReadEncryptedFileRaw()
static DWORD WINAPI send_efs_data(PBYTE pbData,
                                  PVOID pvCallbackContext,
                                  ULONG ulLength)
{
  b_ctx* bctx = (b_ctx*)pvCallbackContext;
  BareosSocket* sd = bctx->jcr->store_bsock;

  if (ulLength == 0) { return ERROR_SUCCESS; }

  /* See if we can fit the data into the current bctx->rbuf which can hold
   * bctx->rsize bytes. */
  if (ulLength <= (ULONG)bctx->rsize) {
    sd->message_length = ulLength;
    memcpy(bctx->rbuf, pbData, ulLength);
    if (!SendDataToSd(bctx)) { return ERROR_NET_WRITE_FAULT; }
  } else {
    // Need to chunk the data into pieces.
    ULONG offset = 0;

    while (ulLength > 0) {
      sd->message_length = MIN((ULONG)bctx->rsize, ulLength);
      memcpy(bctx->rbuf, pbData + offset, sd->message_length);
      if (!SendDataToSd(bctx)) { return ERROR_NET_WRITE_FAULT; }

      offset += sd->message_length;
      ulLength -= sd->message_length;
    }
  }

  return ERROR_SUCCESS;
}

// Send the content of an Encrypted file on an EFS filesystem.
static inline bool SendEncryptedData(b_ctx& bctx)
{
  bool retval = false;

  if (!p_ReadEncryptedFileRaw) {
    Jmsg0(bctx.jcr, M_FATAL, 0,
          _("Encrypted file but no EFS support functions\n"));
  }

  /* The EFS read function, ReadEncryptedFileRaw(), works in a specific way.
   * You have to give it a function that it calls repeatedly every time the
   * read buffer is filled.
   *
   * So ReadEncryptedFileRaw() will not return until it has read the whole file.
   */
  if (p_ReadEncryptedFileRaw((PFE_EXPORT_FUNC)send_efs_data, &bctx,
                             bctx.ff_pkt->bfd.pvContext)) {
    goto bail_out;
  }
  retval = true;

bail_out:
  return retval;
}
#endif

// read data from bfd into buflen sized chunks.  Submit a shared pointer
// of each of them to each channel.
static void ReadData(std::vector<channel::in<shared_buffer>> channels,
		     BareosFilePacket* bfd,
		     std::size_t buflen)
{
  ssize_t max_readsize = 512 * 1024;
  ssize_t readsize = (max_readsize / buflen) * buflen;
  ASSERT(readsize <= max_readsize);
  ASSERT(readsize % buflen == 0);
  // prevent weird situations where max_readsize < buflen
  if (readsize == 0) { readsize = buflen; }
  for (;;) {
    auto mem = std::shared_ptr<char>(new char[readsize], std::default_delete<char[]>{});
    if (!mem) { break; }
    ssize_t num_read = bread(bfd, mem.get(), readsize);
    // some plugins rely on the fact that bread gets called until
    // it returns 0;  the check num_read < readsize is not sufficient here
    // even if it saves some callback calls.
    if (num_read == 0) { break; }

    char* head = mem.get();
    while (num_read > 0) {
      std::size_t size = std::min(static_cast<std::size_t>(num_read), buflen);
      auto buffer = shared_buffer(mem, size, head);
      for (auto& chan : channels) {
	if (!chan.put(buffer)) {
	  return;
	}
      }
      head += size;
      num_read -= size;
    }
  }
}

static void DigestData(channel::out<shared_buffer> out,
		       DIGEST* digest, DIGEST* signing_digest,
		       std::promise<void> barrier)
{
  struct passer {
    std::promise<void> barrier;

    passer(std::promise<void> barrier) : barrier{std::move(barrier)}
    {}

    ~passer() {
      barrier.set_value();
    }
  };

  // notify the barrier once we finish (for any reason!)
  passer p{std::move(barrier)};

  for (std::optional buf = out.get(); buf; buf = out.get()) {
    // Update checksum if requested
    if (digest) {
      CryptoDigestUpdate(digest, (uint8_t*) buf->data.get(), buf->size);
    }

    // Update signing digest if requested
    if (signing_digest) {
      CryptoDigestUpdate(signing_digest, (uint8_t*)buf->data.get(), buf->size);
    }
  }
}

static void Compress(CompressionContext& ctx,
		     channel::out<shared_buffer>& out,
		     std::mutex& input,
		     channel::in<shared_buffer>& in,
		     std::mutex& output,
		     std::condition_variable& cv,
		     std::int64_t& num_in,
		     std::int64_t& num_out,
		     JobControlRecord* jcr,
		     std::size_t max_compress_len,
		     uint32_t compression)
{
  std::optional<shared_buffer> buf = std::nullopt;
  for(;;) {
    std::int64_t block_num = 0;
    {
      std::unique_lock lock(input);
      buf = out.get();
      block_num = num_in++;
    }

    if (!buf) {
      // input is empty; we can stop now
      break;
    }

    shared_buffer compressed(max_compress_len);
    uint32_t size = 0;
    if (!CompressData(jcr, ctx, compression, buf->data.get(),
		      buf->size, (unsigned char*) compressed.data.get(),
		     max_compress_len,
		     &size)) {
      {
	std::unique_lock lock(output);
	num_out = -1;
      }
      cv.notify_all();
      break;
    }

    compressed.size = size;
    {
      std::unique_lock lock(output);
      cv.wait(lock, [&num_out, block_num]() {
	// if the put fails on another thread it will set num_out to
	// -1.
	return (block_num == num_out) || (num_out == -1);
      });
      if (num_out == -1) { return; }
      if (!in.put(std::move(compressed))) { return; }
      num_out++;
    }
    cv.notify_all();
  }
}

static void SendData(channel::out<shared_buffer> block,
		     std::optional<channel::out<shared_buffer>> compressed,
		     b_ctx* bctx, std::promise<void> barrier) {
  struct passer{
    std::promise<void> barrier;

    ~passer() {
      barrier.set_value();
    }
  };

  passer p{std::move(barrier)};

  BareosSocket* sd = bctx->jcr->store_bsock;
  for (std::optional buf = block.get(); buf; buf = block.get()) {
    memcpy(bctx->rbuf, buf->data.get(), buf->size);
    sd->message_length = buf->size;
    std::optional<shared_buffer> comp = std::nullopt;
    if (compressed) {
      if (comp = compressed->get(); !comp) {
	// todo set jcr to cancelled
	Jmsg(bctx->jcr, M_FATAL, 0, "Unexpected end of compression stream.\n");
	break;
      }
    }
    if (!SendDataToSd(bctx, std::move(comp))) {
      Jmsg(bctx->jcr, M_FATAL, 0, "Error while sending data to storage daemon.\n");
      break;
    }
  }
}

// Send the content of a file on anything but an EFS filesystem.
static inline bool SendPlainData(b_ctx& bctx)
{
  std::size_t actual_buf_count;
  {
    std::size_t buflen = bctx.rsize;
    std::size_t filesize = bctx.ff_pkt->statp.st_size;

    std::size_t maxbuffered = 512 * 1024 * 1024; // 512 MB
    std::size_t max_num_bufs = 80;
    // some plugins report the filesize as 0
    // but still create data.
    // to prevent dead locks by trying to insert a buffer into
    // an empty queue, we always set up at least five buffers!
    std::size_t min_num_bufs = 5;

    // do not buffer for much more than maxbuffered bytes
    // and create at most max_num_bufs buffers and at least min_num_bufs buffers

    std::size_t bufferedsize = std::min(filesize, maxbuffered);
    std::size_t num_buffers  = (bufferedsize + buflen - 1) / buflen;

    actual_buf_count = std::clamp(num_buffers, min_num_bufs, max_num_bufs);
  }

  bool retval = true;

  using sbuf = shared_buffer;

  std::vector<channel::in<sbuf>> sinks;
  std::optional<channel::out<shared_buffer>> compression_out = std::nullopt;

  // setup compressing
  if (BitIsSet(FO_COMPRESS, bctx.ff_pkt->flags)) {
    // connection between the reading thread and the compressing thread
    auto [cr_in, cr_out] = channel::CreateBufferedChannel<sbuf>(actual_buf_count);
    sinks.emplace_back(std::move(cr_in));
    // connection between the compressing thread and the sending thread
    auto [sc_in, sc_out] = channel::CreateBufferedChannel<shared_buffer>(1);
    if (!bctx.jcr->fd_impl->send_ctx->to_compress.put({
	  std::move(cr_out), std::move(sc_in),
	  bctx.jcr, bctx.max_compress_len,
	  bctx.ff_pkt->Compress_algo,
	  bctx.ff_pkt->Compress_level})) {
      return false;
    }
    compression_out = std::move(sc_out);
  }

  // setup sending
  std::promise<void> send_barrier{};
  std::future<void>  send_future = send_barrier.get_future();
  {
    // connection between the reading thread and the sending thread
    auto [sr_in, sr_out] = channel::CreateBufferedChannel<sbuf>(actual_buf_count);
    sinks.emplace_back(std::move(sr_in));
    if (!bctx.jcr->fd_impl->send_ctx->to_send.put({
	  std::move(sr_out),
	  std::move(compression_out),
	  &bctx, std::move(send_barrier)})) {
      return false;
    }
  }

  // setup digesting
  std::promise<void> digest_barrier{};
  std::future<void>  digest_future = digest_barrier.get_future();
  {
    // connection between the reading thread and the digesting thread
    auto [dr_in, dr_out] = channel::CreateBufferedChannel<sbuf>(actual_buf_count);
    sinks.emplace_back(std::move(dr_in));
    if (!bctx.jcr->fd_impl->send_ctx->to_digest.put({std::move(dr_out),
	  bctx.digest, bctx.signing_digest, std::move(digest_barrier)})) {
      return false;
    }
  }

  DIGEST* digest = nullptr;
  DIGEST* signing = nullptr;

  // disable digesting inside SendDataToSd
  std::swap(digest, bctx.digest);
  std::swap(signing, bctx.signing_digest);

  ReadData(std::move(sinks), &bctx.ff_pkt->bfd, (std::size_t) bctx.rsize);
  send_future.wait();
  digest_future.wait();
  std::swap(digest, bctx.digest);
  std::swap(signing, bctx.signing_digest);

  return retval;
}

/**
 * Send data read from an already open file descriptor.
 *
 * We return 1 on sucess and 0 on errors.
 *
 * ***FIXME***
 * We use ff_pkt->statp.st_size when FO_SPARSE to know when to stop reading.
 * Currently this is not a problem as the only other stream, resource forks,
 * are not handled as sparse files.
 */
static int send_data(JobControlRecord* jcr,
                     int stream,
                     FindFilesPacket* ff_pkt,
                     DIGEST* digest,
                     DIGEST* signing_digest)
{
  b_ctx bctx;
  BareosSocket* sd = jcr->store_bsock;
#ifdef FD_NO_SEND_TEST
  return 1;
#endif

  // Setup backup context.
  memset(&bctx, 0, sizeof(b_ctx));
  bctx.jcr = jcr;
  bctx.ff_pkt = ff_pkt;
  bctx.msgsave = sd->msg;                  /* save the original sd buffer */
  bctx.rbuf = sd->msg;                     /* read buffer */
  bctx.wbuf = sd->msg;                     /* write buffer */
  bctx.rsize = jcr->buf_size;              /* read buffer size */
  bctx.cipher_input = (uint8_t*)bctx.rbuf; /* encrypt uncompressed data */
  bctx.digest = digest;                    /* encryption digest */
  bctx.signing_digest = signing_digest;    /* signing digest */

  Dmsg1(300, "Saving data, type=%d\n", ff_pkt->type);

  if (!SetupCompressionContext(bctx)) { goto bail_out; }

  if (!SetupEncryptionContext(bctx)) { goto bail_out; }

  /* Send Data header to Storage daemon
   *    <file-index> <stream> <info> */
  if (!sd->fsend("%ld %d 0", jcr->JobFiles, stream)) {
    if (!jcr->IsJobCanceled()) {
      Jmsg1(jcr, M_FATAL, 0, _("Network send error to SD. ERR=%s\n"),
            sd->bstrerror());
    }
    goto bail_out;
  }
  Dmsg1(300, ">stored: datahdr %s", sd->msg);

  /* Make space at beginning of buffer for fileAddr because this
   *   same buffer will be used for writing if compression is off. */
  if (BitIsSet(FO_SPARSE, ff_pkt->flags)
      || BitIsSet(FO_OFFSETS, ff_pkt->flags)) {
    bctx.rbuf += OFFSET_FADDR_SIZE;
    bctx.rsize -= OFFSET_FADDR_SIZE;
#ifdef HAVE_FREEBSD_OS
    // To read FreeBSD partitions, the read size must be a multiple of 512.
    bctx.rsize = (bctx.rsize / 512) * 512;
#endif
  }

  // A RAW device read on win32 only works if the buffer is a multiple of 512
#ifdef HAVE_WIN32
  if (S_ISBLK(ff_pkt->statp.st_mode)) { bctx.rsize = (bctx.rsize / 512) * 512; }

  if (ff_pkt->statp.st_rdev & FILE_ATTRIBUTE_ENCRYPTED) {
    if (!SendEncryptedData(bctx)) { goto bail_out; }
  } else {
    if (!SendPlainData(bctx)) { goto bail_out; }
  }
#else
  if (!SendPlainData(bctx)) { goto bail_out; }
#endif

  if (sd->message_length < 0) { /* error */
    BErrNo be;
    Jmsg(jcr, M_ERROR, 0, _("Read error on file %s. ERR=%s\n"), ff_pkt->fname,
         be.bstrerror(ff_pkt->bfd.BErrNo));
    if (jcr->JobErrors++ > 1000) { /* insanity check */
      Jmsg(jcr, M_FATAL, 0, _("Too many errors. JobErrors=%d.\n"),
           jcr->JobErrors);
    }
  } else if (BitIsSet(FO_ENCRYPT, ff_pkt->flags)) {
    // For encryption, we must call finalize to push out any buffered data.
    if (!CryptoCipherFinalize(bctx.cipher_ctx,
                              (uint8_t*)jcr->fd_impl->crypto.crypto_buf,
                              &bctx.encrypted_len)) {
      // Padding failed. Shouldn't happen.
      Jmsg(jcr, M_FATAL, 0, _("Encryption padding error\n"));
      goto bail_out;
    }

    // Note, on SSL pre-0.9.7, there is always some output
    if (bctx.encrypted_len > 0) {
      sd->message_length = bctx.encrypted_len;   /* set encrypted length */
      sd->msg = jcr->fd_impl->crypto.crypto_buf; /* set correct write buffer */
      if (!sd->send()) {
        if (!jcr->IsJobCanceled()) {
          Jmsg1(jcr, M_FATAL, 0, _("Network send error to SD. ERR=%s\n"),
                sd->bstrerror());
        }
        goto bail_out;
      }
      Dmsg1(130, "Send data to SD len=%d\n", sd->message_length);
      jcr->JobBytes += sd->message_length; /* count bytes saved possibly
                                              compressed/encrypted */
      sd->msg = bctx.msgsave;              /* restore bnet buffer */
    }
  }

  if (!sd->signal(BNET_EOD)) { /* indicate end of file data */
    if (!jcr->IsJobCanceled()) {
      Jmsg1(jcr, M_FATAL, 0, _("Network send error to SD. ERR=%s\n"),
            sd->bstrerror());
    }
    goto bail_out;
  }

  // Free the cipher context
  if (bctx.cipher_ctx) { CryptoCipherFree(bctx.cipher_ctx); }

  return 1;

bail_out:
  // Free the cipher context
  if (bctx.cipher_ctx) { CryptoCipherFree(bctx.cipher_ctx); }

  sd->msg = bctx.msgsave; /* restore bnet buffer */
  sd->message_length = 0;

  return 0;
}

static bool SendFileHeader(
    BareosSocket* sd,
    int32_t file_index,  // attention: is 32bit, gets send with %ld
    int type,
    const char* canonical_name,
    const char* encoded_attributes,
    const char* link_name,
    const char* encoded_ex_attributes,
    int32_t delta_seq  // attention: is signed, gets send as unsigned
)
{
  // fsend already would already do these substitutions for us,
  // but this makes it clearer.
  if (!canonical_name) canonical_name = "";
  if (!encoded_attributes) encoded_attributes = "";
  if (!link_name) link_name = "";
  if (!encoded_ex_attributes) encoded_ex_attributes = "";
  bool status = sd->fsend("%ld %d %s%c%s%c%s%c%s%c%u%c", file_index, type,
                          canonical_name, 0, encoded_attributes, 0, link_name,
                          0, encoded_ex_attributes, 0, delta_seq, 0);
  return status;
}

static bool SendRestoreObject(BareosSocket* sd,
                              int32_t file_index,
                              int file_type,
                              const char* file_name,
                              const char* object_name,
                              int32_t object_index,
                              int32_t object_len,
                              char* object_data)
{
  int comp_len = object_len;
  bool object_compression = 0;

  char* send_data = object_data;
  PoolMem compressed_object;

  if (object_len > 1000) {
    // Big object, compress it
    comp_len = compressBound(object_len);
    // FIXME: check Zdeflate error
    Zdeflate((char*)object_data, object_len,
             compressed_object.check_size(comp_len), comp_len);
    if (comp_len < object_len) {
      // compressed object is smaller; use it
      send_data = compressed_object.c_str();
      object_compression = 1; /* zlib level 9 compression */
      Dmsg2(100, "Object compressed from %d to %d bytes\n", object_len,
            comp_len);
    }
  }

  sd->message_length = Mmsg(sd->msg, "%d %d %d %d %d %d %s%c%s%c", file_index,
                            file_type, object_index, comp_len, object_len,
                            object_compression, file_name, 0, object_name, 0);
  sd->msg = CheckPoolMemorySize(sd->msg, sd->message_length + comp_len + 2);
  memcpy(sd->msg + sd->message_length, send_data, comp_len);

  // Note we send one extra byte so Dir can store zero after object
  sd->message_length += comp_len + 1;
  return sd->send();
}

static const char* GetCanonicalName(const struct stat& statp,
                                    const char* fname,
                                    const char* link)
{
  const char* selected;
  if (S_ISDIR(statp.st_mode)) {
    selected = link;
  } else {
    selected = fname;
  }

  return selected;
}

static const char* GetLinkName(int type,
                               const char* fname [[maybe_unused]],
                               const char* link)
{
  const char* selected;
  switch (type) {
    case FT_JUNCTION:
    case FT_LNK:
    case FT_LNKSAVED: {
      selected = link;
    } break;
    default: {
      // should we return nullptr here instead ?
      selected = link;
    } break;
  }

  return selected;
}

bool EncodeAndSendAttributes(JobControlRecord* jcr,
                             FindFilesPacket* ff_pkt,
                             int& data_stream)
{
  BareosSocket* sd = jcr->store_bsock;
  PoolMem attribs(PM_NAME), attribsExBuf(PM_NAME);
  char* attribsEx = NULL;
  int attr_stream;
  bool status;
  int hangup = GetHangup();
#ifdef FD_NO_SEND_TEST
  return true;
#endif

  Dmsg1(300, "encode_and_send_attrs fname=%s\n", ff_pkt->fname);
  /** Find what data stream we will use, then encode the attributes */
  if ((data_stream = SelectDataStream(ff_pkt)) == STREAM_NONE) {
    /* This should not happen */
    Jmsg0(jcr, M_FATAL, 0,
          _("Invalid file flags, no supported data stream type.\n"));
    return false;
  }
  EncodeStat(attribs.c_str(), &ff_pkt->statp, sizeof(ff_pkt->statp),
             ff_pkt->LinkFI, data_stream);

  /** Now possibly extend the attributes */
  if (IS_FT_OBJECT(ff_pkt->type)) {
    attr_stream = STREAM_RESTORE_OBJECT;
  } else {
    attribsEx = attribsExBuf.c_str();
    attr_stream = encode_attribsEx(jcr, attribsEx, ff_pkt);
  }

  Dmsg3(300, "File %s\nattribs=%s\nattribsEx=%s\n", ff_pkt->fname,
        attribs.c_str(), attribsEx);

  jcr->lock();
  ff_pkt->FileIndex = ++jcr->JobFiles; /* return FileIndex */
  PmStrcpy(jcr->fd_impl->last_fname, ff_pkt->fname);
  jcr->unlock();

  // Debug code: check if we must hangup
  if (hangup && (ff_pkt->FileIndex > hangup)) {
    jcr->setJobStatusWithPriorityCheck(JS_Incomplete);
    Jmsg1(jcr, M_FATAL, 0, "Debug hangup requested after %d files.\n", hangup);
    SetHangup(0);
    return false;
  }

  /* Send Attributes header to Storage daemon
   *    <file-index> <stream> <info> */
  if (!sd->fsend("%ld %d 0", ff_pkt->FileIndex, attr_stream)) {
    if (!jcr->IsJobCanceled() && !jcr->IsIncomplete()) {
      Jmsg1(jcr, M_FATAL, 0, _("Network send error to SD. ERR=%s\n"),
            sd->bstrerror());
    }
    return false;
  }
  Dmsg1(300, ">stored: attrhdr %s", sd->msg);

  /* Send file attributes to Storage daemon
   *   File_index
   *   File type
   *   Filename (full path)
   *   Encoded attributes
   *   Link name (if type==FT_LNK or FT_LNKSAVED)
   *   Encoded extended-attributes (for Win32)
   *   Delta Sequence Number
   *
   * or send Restore Object to Storage daemon
   *   File_index
   *   File_type
   *   Object_index
   *   Object_len  (possibly compressed)
   *   Object_full_len (not compressed)
   *   Object_compression
   *   Plugin_name
   *   Object_name
   *   Binary Object data
   *
   * For a directory, link is the same as fname, but with trailing
   * slash. For a linked file, link is the link. */

  std::string StrippedFile, StrippedLink;
  const char* file_name = nullptr;
  const char* link_name = nullptr;
  /* If requested strip leading components of the path so that we can
   * save file as if it came from a subdirectory.  This is most useful
   * for dealing with snapshots, by removing the snapshot directory, or
   * in handling vendor migrations where files have been restored with
   * a vendor product into a subdirectory. */
  bool do_strip_link = (ff_pkt->type != FT_LNK) && (ff_pkt->fname != ff_pkt->link);
  if (!IS_FT_OBJECT(ff_pkt->type) && ff_pkt->type != FT_DELETED
      && ShouldStripPaths(ff_pkt)) { /* already stripped */
    bool file_stripped
        = do_strip(ff_pkt->StripPath, ff_pkt->fname, StrippedFile);

    if (do_strip_link)
    {
	    bool link_stripped
		    = do_strip(ff_pkt->StripPath, ff_pkt->link, StrippedLink);

	    if (file_stripped || link_stripped) {
		    file_name = GetCanonicalName(ff_pkt->statp, StrippedFile.c_str(),
						 StrippedLink.c_str());
		    link_name = GetLinkName(ff_pkt->type, StrippedFile.c_str(),
					    StrippedLink.c_str());
		    Dmsg3(100, "fname=%s stripped=%s link=%s\n", ff_pkt->fname,
			  StrippedFile.c_str(), StrippedLink.c_str());
	    }
    }
    else
    {
	    if (file_stripped)
	    {
		    file_name = GetCanonicalName(ff_pkt->statp, StrippedFile.c_str(),
						 ff_pkt->link);
		    link_name = GetLinkName(ff_pkt->type, StrippedFile.c_str(),
					    ff_pkt->link);
		    Dmsg3(100, "fname=%s stripped=%s link=%s\n", ff_pkt->fname,
			  StrippedFile.c_str(), StrippedLink.c_str());

	    }
    }

  }

  if (!file_name) {
    file_name = GetCanonicalName(ff_pkt->statp, ff_pkt->fname, ff_pkt->link);
    link_name = GetLinkName(ff_pkt->type, ff_pkt->fname, ff_pkt->link);
  }

  if ((ff_pkt->type == FT_PLUGIN_CONFIG)
      || (ff_pkt->type == FT_RESTORE_FIRST)) {
    status = SendRestoreObject(sd, ff_pkt->FileIndex, ff_pkt->type, file_name,
                               ff_pkt->object_name, ff_pkt->object_index,
                               ff_pkt->object_len, ff_pkt->object);
  } else {
    status = SendFileHeader(sd, ff_pkt->FileIndex, ff_pkt->type, file_name,
                            attribs.c_str(), link_name, attribsEx,
                            ff_pkt->delta_seq);
  }

  switch (ff_pkt->type) {
    case FT_JUNCTION:
    case FT_LNK:
    case FT_LNKSAVED: {
      Dmsg3(300, "Link %d %s to %s\n", ff_pkt->FileIndex, file_name, link_name);
    } break;
  }

  Dmsg2(300, ">stored: attr len=%d: %s\n", sd->message_length, sd->msg);
  if (!status && !jcr->IsJobCanceled()) {
    Jmsg1(jcr, M_FATAL, 0, _("Network send error to SD. ERR=%s\n"),
          sd->bstrerror());
  }

  sd->signal(BNET_EOD); /* indicate end of attributes data */

  return status;
}

// given a path and a number of path segments to skip
// this function returns a copy of the given path, except it removes
// the segments 2, 3, 4, ..., n, where n was the given number.
// example: (/a/b/c/d, 2) -> (/c/d); think of the path as [/; a/; b/; c/; d]
//          (C:\Windows\Test, 1) -> (C:\Test); ~> [C:\; Windows\; Test\]
// if n is bigger than the number of segments in the path then only the
// first segment is copied
bool do_strip(int count, const char* in, std::string& out)
{
  int stripped;
  int numsep = 0;

  // Copy to first path separator -- Win32 might have c: ...
  while (*in && !IsPathSeparator(*in)) { out.push_back(*in++); }
  if (*in) { /* Not at the end of the string */
	  out.push_back(*in++);
    numsep++; /* one separator seen */
  }
  for (stripped = 0; stripped < count && *in; stripped++) {
    while (*in && !IsPathSeparator(*in)) { in++; /* skip chars */ }
    if (*in) {
      numsep++; /* count separators seen */
      in++;     /* skip separator */
    }
  }

  // Copy to end
  while (*in) { /* copy to end */
    if (IsPathSeparator(*in)) { numsep++; }
    out.push_back(*in++);
  }
  Dmsg4(500, "stripped=%d count=%d numsep=%d sep>count=%d\n", stripped, count,
        numsep, numsep > count);
  return stripped == count && numsep > count;
}

bool ShouldStripPaths(const FindFilesPacket* ff_pkt)
{
  if (!BitIsSet(FO_STRIPPATH, ff_pkt->flags) || ff_pkt->StripPath <= 0) {
    Dmsg1(200, "No strip for %s\n", ff_pkt->fname);
    return false;
  }
  return true;
}

std::string GetStrippedCanonicalName(const FindFilesPacket* ff_pkt)
{
	std::string stripped_name;

  const char* name
      = GetCanonicalName(ff_pkt->statp, ff_pkt->fname, ff_pkt->link);

  if (ShouldStripPaths(ff_pkt)) {
    do_strip(ff_pkt->StripPath, name, stripped_name);
  } else {
    stripped_name = std::string{name};
  }

  return stripped_name;
}

#if defined(WIN32_VSS)
static void CloseVssBackupSession(JobControlRecord* jcr)
{
  /* STOP VSS ON WIN32
   * Tell vss to close the backup session */
  if (jcr->fd_impl->pVSSClient) {
    /* We are about to call the BackupComplete VSS method so let all plugins
     * know that by raising the bEventVssBackupComplete event. */
    GeneratePluginEvent(jcr, bEventVssBackupComplete);
    if (jcr->fd_impl->pVSSClient->CloseBackup()) {
      // Inform user about writer states
      for (size_t i = 0; i < jcr->fd_impl->pVSSClient->GetWriterCount(); i++) {
        int msg_type = M_INFO;
        if (jcr->fd_impl->pVSSClient->GetWriterState(i) < 1) {
          msg_type = M_WARNING;
          jcr->JobErrors++;
        }
        Jmsg(jcr, msg_type, 0, _("VSS Writer (BackupComplete): %s\n"),
             jcr->fd_impl->pVSSClient->GetWriterInfo(i));
      }
    }

    // Generate Job global writer metadata
    wchar_t* metadata = jcr->fd_impl->pVSSClient->GetMetadata();
    if (metadata) {
      FindFilesPacket* ff_pkt = jcr->fd_impl->ff;
      ff_pkt->fname = (char*)"*all*"; /* for all plugins */
      ff_pkt->type = FT_RESTORE_FIRST;
      ff_pkt->LinkFI = 0;
      ff_pkt->object_name = (char*)"job_metadata.xml";
      ff_pkt->object = BSTR_2_str(metadata);
      ff_pkt->object_len = (wcslen(metadata) + 1) * sizeof(wchar_t);
      ff_pkt->object_index = (int)time(NULL);
      SaveFile(jcr, ff_pkt, true);
    }
  }
}
#endif
} /* namespace filedaemon */
