/*
   BAREOS® - Backup Archiving REcovery Open Sourced

   Copyright (C) 2000-2012 Free Software Foundation Europe e.V.
   Copyright (C) 2011-2012 Planets Communications B.V.
   Copyright (C) 2013-2023 Bareos GmbH & Co. KG

   This program is Free Software; you can redistribute it and/or
   modify it under the terms of version three of the GNU Affero General Public
   License as published by the Free Software Foundation and included
   in the file LICENSE.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Affero General Public License for more details.

   You should have received a copy of the GNU Affero General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.
*/
// Kern E. Sibbald, MM
/**
 * @file
 * Main routine for finding files on a file system.
 * The heart of the work to find the files on the
 * system is done in find_one.c. Here we have the
 * higher level control as well as the matching
 * routines for the new syntax Options resource.
 */

#include <unistd.h>
#include "include/bareos.h"
#include "include/filetypes.h"
#include "include/jcr.h"
#include "find.h"
#include "findlib/find_one.h"
#include "lib/util.h"
#include "lib/berrno.h"
#include "lib/thread_specific_data.h"
#include "hardlink.h"

#include <memory>  // unique_ptr
#include <initializer_list>
#include <utility>  // pair
#include <thread>
#include <future>

#if defined(HAVE_DARWIN_OS)
/* the MacOS linker wants symbols for the destructors of these two types, so we
 * have to force template instantiation. */
template class alist<findFOPTS*>;
template class dlist<dlistString>;
#endif

static const int debuglevel = 450;

int32_t name_max; /* filename max length */
int32_t path_max; /* path name max length */

static const int fnmode = 0;

// Initialize the find files "global" variables
FindFilesPacket* init_find_files()
{
  FindFilesPacket* ff = new (malloc(sizeof(FindFilesPacket))) FindFilesPacket;

  /* Get system path and filename maximum lengths */
  path_max = pathconf(".", _PC_PATH_MAX);
  if (path_max < 2048) { path_max = 2048; }

  name_max = pathconf(".", _PC_NAME_MAX);
  if (name_max < 2048) { name_max = 2048; }
  path_max++; /* add for EOS */
  name_max++; /* add for EOS */

  Dmsg1(debuglevel, "init_find_files ff=%p\n", ff);
  return ff;
}

/**
 * Set FindFiles options. For the moment, we only
 * provide for full/incremental saves, and setting
 * of save_time. For additional options, see above
 */
void SetFindOptions(FindFilesPacket* ff, bool incremental, time_t save_time)
{
  Dmsg0(debuglevel, "Enter SetFindOptions()\n");
  ff->incremental = incremental;
  ff->save_time = save_time;
  Dmsg0(debuglevel, "Leave SetFindOptions()\n");
}

void SetFindChangedFunction(FindFilesPacket* ff,
                            bool CheckFct(JobControlRecord* jcr,
                                          FindFilesPacket* ff))
{
  Dmsg0(debuglevel, "Enter SetFindChangedFunction()\n");
  ff->CheckFct = CheckFct;
}

static void SetupLastOptionBlock(FindFilesPacket* ff,
                                 findIncludeExcludeItem* incexe)
{
  /* By setting all options, we in effect OR the global options which is
   * what we want. */
  findFOPTS* fo = incexe->opts_list.get(incexe->opts_list.size() - 1);
  CopyBits(FO_MAX, fo->flags, ff->flags);
  ff->incexe = incexe;
  ff->Compress_algo = fo->Compress_algo;
  ff->Compress_level = fo->Compress_level;
  ff->StripPath = fo->StripPath;
  ff->size_match = fo->size_match;
  ff->fstypes = &fo->fstype;
  ff->drivetypes = &fo->Drivetype;


  // reset plugins
  ff->plugin = NULL;
  ff->opt_plugin = false;
  for (int j = 0; j < incexe->opts_list.size(); j++) {
    findFOPTS* fo = (findFOPTS*)incexe->opts_list.get(j);
    if (fo->plugin != NULL) {
      ff->plugin = fo->plugin; /* TODO: generate a plugin event ? */
      ff->opt_plugin = true;
    }
  }

  // reset opts
  strcpy(ff->VerifyOpts, "V");
  strcpy(ff->AccurateOpts, "Cmcs");  /* mtime+ctime+size by default */
  strcpy(ff->BaseJobOpts, "Jspug5"); /* size+perm+user+group+chk  */
  for (int j = 0; j < incexe->opts_list.size(); j++) {
    findFOPTS* fo = (findFOPTS*)incexe->opts_list.get(j);
    if (fo->AccurateOpts[0]) {
      bstrncpy(ff->AccurateOpts, fo->AccurateOpts, sizeof(ff->AccurateOpts));
    }
    if (fo->BaseJobOpts[0]) {
      bstrncpy(ff->BaseJobOpts, fo->BaseJobOpts, sizeof(ff->BaseJobOpts));
    }
    bstrncat(ff->VerifyOpts, fo->VerifyOpts,
             sizeof(ff->VerifyOpts)); /* TODO: Concat or replace? */
  }
}

/**
 * The code comes here for each file examined.
 * We filter the files, then call the user's callback if the file is included.
 */
#include <functional>
static auto CreateCallback(std::function<int(JobControlRecord* jcr,
                                             FindFilesPacket* ff,
                                             bool top_level)> SaveFile)
{
  return
      [SaveFile](JobControlRecord* jcr, FindFilesPacket* ff, bool top_level) {
        if (top_level) {
          return SaveFile(jcr, ff, top_level); /* accept file */
        }
        switch (ff->type) {
          case FT_NOACCESS:
          case FT_NOFOLLOW:
          case FT_NOSTAT:
          case FT_NOCHG:
          case FT_ISARCH:
          case FT_NORECURSE:
          case FT_NOFSCHG:
          case FT_INVALIDFS:
          case FT_INVALIDDT:
          case FT_NOOPEN:
            //    return ff->FileSave(jcr, ff, top_level);

            /* These items can be filtered */
          case FT_LNKSAVED:
          case FT_REGE:
          case FT_REG:
          case FT_LNK:
          case FT_DIRBEGIN:
          case FT_DIREND:
          case FT_RAW:
          case FT_FIFO:
          case FT_SPEC:
          case FT_DIRNOCHG:
          case FT_REPARSE:
          case FT_JUNCTION:
            if (AcceptFile(ff)) {
              return SaveFile(jcr, ff, top_level);
            } else {
              Dmsg1(debuglevel, "Skip file %s\n", ff->fname);
              return -1; /* ignore this file */
            }

          default:
            Dmsg1(000, "Unknown FT code %d\n", ff->type);
            return 0;
        }
      };
}


/**
 * Call this subroutine with a callback subroutine as the first
 * argument and a packet as the second argument, this packet
 * will be passed back to the callback subroutine as the last
 * argument.
 */
int FindFiles(JobControlRecord* jcr,
              FindFilesPacket* ff,
              int FileSave(JobControlRecord* jcr,
                           FindFilesPacket* ff_pkt,
                           bool top_level),
              int PluginSave(JobControlRecord* jcr,
                             FindFilesPacket* ff_pkt,
                             bool top_level))
{
  /* This is the new way */
  findFILESET* fileset = ff->fileset;
  if (fileset) {
    /* TODO: We probably need be move the initialization in the fileset loop,
     * at this place flags options are "concatenated" accross Include {} blocks
     * (not only Options{} blocks inside a Include{}) */
    ClearAllBits(FO_MAX, ff->flags);
    for (int i = 0; i < fileset->include_list.size(); i++) {
      dlistString* node;
      findIncludeExcludeItem* incexe
          = (findIncludeExcludeItem*)fileset->include_list.get(i);
      SetupLastOptionBlock(ff, incexe);

      Dmsg4(50, "Verify=<%s> Accurate=<%s> BaseJob=<%s> flags=<%d>\n",
            ff->VerifyOpts, ff->AccurateOpts, ff->BaseJobOpts, ff->flags);

      foreach_dlist (node, &incexe->name_list) {
        char* fname = node->c_str();

        Dmsg1(debuglevel, "F %s\n", fname);
        ff->top_fname = fname;
        if (FindOneFile(jcr, ff, CreateCallback(FileSave), ff->top_fname,
                        (dev_t)-1, true)
            == 0) {
          return 0; /* error return */
        }
        if (jcr->IsJobCanceled()) { return 0; }
      }

      foreach_dlist (node, &incexe->plugin_list) {
        char* fname = node->c_str();

        if (!PluginSave) {
          Jmsg(jcr, M_FATAL, 0, _("Plugin: \"%s\" not found.\n"), fname);
          return 0;
        }
        Dmsg1(debuglevel, "PluginCommand: %s\n", fname);
        ff->top_fname = fname;
        ff->cmd_plugin = true;
        PluginSave(jcr, ff, true);
        ff->cmd_plugin = false;
        if (jcr->IsJobCanceled()) { return 0; }
      }
    }
  }
  return 1;
}

/**
 * Test if the currently selected directory (in ff->fname) is
 * explicitly in the Include list or explicitly in the Exclude list.
 */
bool IsInFileset(FindFilesPacket* ff)
{
  int i;
  char* fname;
  dlistString* node;
  findIncludeExcludeItem* incexe;
  findFILESET* fileset = ff->fileset;

  if (fileset) {
    for (i = 0; i < fileset->include_list.size(); i++) {
      incexe = (findIncludeExcludeItem*)fileset->include_list.get(i);
      foreach_dlist (node, &incexe->name_list) {
        fname = node->c_str();
        Dmsg2(debuglevel, "Inc fname=%s ff->fname=%s\n", fname, ff->fname);
        if (bstrcmp(fname, ff->fname)) { return true; }
      }
    }
    for (i = 0; i < fileset->exclude_list.size(); i++) {
      incexe = (findIncludeExcludeItem*)fileset->exclude_list.get(i);
      foreach_dlist (node, &incexe->name_list) {
        fname = node->c_str();
        Dmsg2(debuglevel, "Exc fname=%s ff->fname=%s\n", fname, ff->fname);
        if (bstrcmp(fname, ff->fname)) { return true; }
      }
    }
  }

  return false;
}

static std::optional<const char*> FindMatch(
    int (*match_func)(const char* pattern, const char* str, int flags),
    const std::vector<std::string>& patterns,
    const char* str,
    int flags)
{
  for (auto& pattern : patterns) {
    if (match_func(pattern.c_str(), str, flags) == 0) {
      return std::make_optional(pattern.c_str());
    }
  }
  return std::nullopt;
}

static std::optional<const regex_t*> FindRegexMatch(
    std::vector<regex_t>& regexs,
    const char* str)
{
  for (auto& regex : regexs) {
    if (regexec(&regex, str, 0, NULL, 0) == 0) {
      return std::make_optional(&regex);
    }
  }
  return std::nullopt;
}

bool AcceptFile(FindFilesPacket* ff)
{
  const char* basename;
  findFILESET* fileset = ff->fileset;
  findIncludeExcludeItem* incexe = ff->incexe;
  int (*match_func)(const char* pattern, const char* string, int flags);

  Dmsg1(debuglevel, "enter AcceptFile: fname=%s\n", ff->fname);
  if (BitIsSet(FO_ENHANCEDWILD, ff->flags)) {
    match_func = fnmatch;
    if ((basename = last_path_separator(ff->fname)) != NULL)
      basename++;
    else
      basename = ff->fname;
  } else {
    match_func = fnmatch;
    basename = ff->fname;
  }

  for (int j = 0; j < incexe->opts_list.size(); j++) {
    findFOPTS* fo = (findFOPTS*)incexe->opts_list.get(j);
    CopyBits(FO_MAX, fo->flags, ff->flags);
    ff->Compress_algo = fo->Compress_algo;
    ff->Compress_level = fo->Compress_level;
    ff->fstypes = &fo->fstype;
    ff->drivetypes = &fo->Drivetype;

    const int fnm_flags
        = (BitIsSet(FO_IGNORECASE, ff->flags) ? FNM_CASEFOLD : 0)
          | (BitIsSet(FO_ENHANCEDWILD, ff->flags) ? FNM_PATHNAME : 0);

    bool do_exclude = BitIsSet(FO_EXCLUDE, ff->flags);

    if (S_ISDIR(ff->statp.st_mode)) {
      // wild matches
      if (std::optional match
          = FindMatch(match_func, fo->wilddir, ff->fname, fnmode | fnm_flags);
          match.has_value()) {
        if (do_exclude) {
          Dmsg2(debuglevel, "Exclude wilddir: %s file=%s\n", match.value(),
                ff->fname);
        }
        return !do_exclude;
      }
      if (std::optional match
          = FindMatch(match_func, fo->wild, ff->fname, fnmode | fnm_flags);
          match.has_value()) {
        if (do_exclude) {
          Dmsg2(debuglevel, "Exclude wild: %s file=%s\n", match.value(),
                ff->fname);
        }
        return !do_exclude;
      }
      // regex matches
      if (std::optional match = FindRegexMatch(fo->regexdir, ff->fname);
          match.has_value()) {
        if (do_exclude) {
          Dmsg2(debuglevel, "Exclude regexdir: file=%s\n", ff->fname);
        }
        return !do_exclude;
      }
      if (std::optional match = FindRegexMatch(fo->regex, ff->fname);
          match.has_value()) {
        if (do_exclude) {
          Dmsg2(debuglevel, "Exclude regex: file=%s\n", ff->fname);
        }
        return !do_exclude;
      }
    } else {
      // wild matches
      if (std::optional match
          = FindMatch(match_func, fo->wildfile, ff->fname, fnmode | fnm_flags);
          match.has_value()) {
        if (do_exclude) {
          Dmsg2(debuglevel, "Exclude wildfile: %s file=%s\n", match.value(),
                ff->fname);
        }
        return !do_exclude;
      }
      if (std::optional match
          = FindMatch(match_func, fo->wildbase, basename, fnmode | fnm_flags);
          match.has_value()) {
        if (do_exclude) {
          Dmsg2(debuglevel, "Exclude wildbase: %s file=%s\n", match.value(),
                ff->fname);
        }
        return !do_exclude;
      }
      if (std::optional match
          = FindMatch(match_func, fo->wild, ff->fname, fnmode | fnm_flags);
          match.has_value()) {
        if (do_exclude) {
          Dmsg2(debuglevel, "Exclude wild: %s file=%s\n", match.value(),
                ff->fname);
        }
        return !do_exclude;
      }
      // regex matches
      if (std::optional match = FindRegexMatch(fo->regexfile, ff->fname);
          match.has_value()) {
        if (do_exclude) {
          Dmsg2(debuglevel, "Exclude regexfile: file=%s\n", ff->fname);
        }
        return !do_exclude;
      }
      if (std::optional match = FindRegexMatch(fo->regex, ff->fname);
          match.has_value()) {
        if (do_exclude) {
          Dmsg2(debuglevel, "Exclude regex: file=%s\n", ff->fname);
        }
        return !do_exclude;
      }
    }

    // If we have an empty Options clause with exclude, then exclude the file
    if (do_exclude && fo->regex.size() == 0 && fo->wild.size() == 0
        && fo->regexdir.size() == 0 && fo->wilddir.size() == 0
        && fo->regexfile.size() == 0 && fo->wildfile.size() == 0
        && fo->wildbase.size() == 0) {
      Dmsg1(debuglevel, "Empty options, rejecting: %s\n", ff->fname);
      return false; /* reject file */
    }
  }

  // Now apply the Exclude { } directive
  for (int i = 0; i < fileset->exclude_list.size(); i++) {
    dlistString* node;
    findIncludeExcludeItem* incexe
        = (findIncludeExcludeItem*)fileset->exclude_list.get(i);

    for (int j = 0; j < incexe->opts_list.size(); j++) {
      findFOPTS* fo = (findFOPTS*)incexe->opts_list.get(j);
      const int fnm_flags
          = BitIsSet(FO_IGNORECASE, fo->flags) ? FNM_CASEFOLD : 0;
      if (std::optional match
          = FindMatch(fnmatch, fo->wild, ff->fname, fnmode | fnm_flags);
          match.has_value()) {
        Dmsg2(debuglevel, "Reject wild1: %s\n", ff->fname);
        return false; /* reject file */
      }
    }
    const int fnm_flags
        = (incexe->current_opts != NULL
           && BitIsSet(FO_IGNORECASE, incexe->current_opts->flags))
              ? FNM_CASEFOLD
              : 0;
    foreach_dlist (node, &incexe->name_list) {
      char* fname = node->c_str();

      if (fnmatch(fname, ff->fname, fnmode | fnm_flags) == 0) {
        Dmsg1(debuglevel, "Reject wild2: %s\n", ff->fname);
        return false; /* reject file */
      }
    }
  }

  return true;
}

// Terminate FindFiles() and release all allocated memory
void TermFindFiles(FindFilesPacket* ff)
{
  if (ff) {
    TermFindOne(ff);
    std::destroy_at(ff);
    free(ff);
  }
}

// Allocate a new include/exclude block.
findIncludeExcludeItem* allocate_new_incexe(void)
{
  findIncludeExcludeItem* incexe
      = (findIncludeExcludeItem*)malloc(sizeof(findIncludeExcludeItem));
  new (incexe) findIncludeExcludeItem{};

  return incexe;
}

// Define a new Exclude block in the FileSet
findIncludeExcludeItem* new_exclude(FindFilesPacket* ff)
{
  // New exclude
  ff->incexe = allocate_new_incexe();
  ff->fileset->exclude_list.append(ff->incexe);

  return ff->incexe;
}

// Define a new Include block in the FileSet
findIncludeExcludeItem* new_include(FindFilesPacket* ff)
{
  // New include
  ff->incexe = allocate_new_incexe();
  ff->fileset->include_list.append(ff->incexe);

  return ff->incexe;
}

/**
 * Define a new preInclude block in the FileSet.
 * That is the include is prepended to the other
 * Includes. This is used for plugin exclusions.
 */
findIncludeExcludeItem* new_preinclude(FindFilesPacket* ff)
{
  // New pre-include
  ff->incexe = allocate_new_incexe();
  ff->fileset->include_list.prepend(ff->incexe);

  return ff->incexe;
}

findFOPTS* start_options(FindFilesPacket* ff)
{
  int state = ff->fileset->state;
  findIncludeExcludeItem* incexe = ff->incexe;

  if (state != state_options) {
    ff->fileset->state = state_options;
    findFOPTS* fo = (findFOPTS*)malloc(sizeof(findFOPTS));
    new (fo) findFOPTS{};
    incexe->current_opts = fo;
    incexe->opts_list.append(fo);
  }

  return incexe->current_opts;
}

// Used by plugins to define a new options block
void NewOptions(FindFilesPacket* ff, findIncludeExcludeItem* incexe)
{
  findFOPTS* fo;

  fo = (findFOPTS*)malloc(sizeof(findFOPTS));
  new (fo) findFOPTS{};
  incexe->current_opts = fo;
  incexe->opts_list.prepend(fo);
  ff->fileset->state = state_options;
}

auto SaveInList(found_files_list& result)
{
  return [&result](JobControlRecord* jcr, FindFilesPacket* ff_pkt, bool) {
    switch (ff_pkt->type) {
      case FT_LNKSAVED: /* Hard linked, file already saved */
        Dmsg2(50, "FT_LNKSAVED hit while listing: %s => %s (ERROR)\n",
              ff_pkt->fname, ff_pkt->link);
        return 0;
        break;
      case FT_REGE:
        Dmsg1(130, "found FT_REGE: %s\n", ff_pkt->fname);
        break;
      case FT_REG:
        Dmsg1(130, "found FT_REG: %s\n", ff_pkt->fname);
        break;
      case FT_LNK:
        Dmsg2(130, "found FT_LNK: %s -> %s\n", ff_pkt->fname, ff_pkt->link);
        break;
      case FT_RESTORE_FIRST:
        Dmsg2(50, "FT_RESTORE_FIRST hit while listing: %s (ERROR)\n",
              ff_pkt->fname);
        return 0;
        break;
      case FT_PLUGIN_CONFIG:
        Dmsg2(50, "FT_PLUGIN_CONFIG hit while listing: %s (ERROR)\n",
              ff_pkt->fname);
        return 0;
        break;
      case FT_DIRBEGIN:
        /* this is skipped, so num_skipped is not increased */
        return 1; /* not used */
      case FT_NORECURSE:
        Jmsg(jcr, M_INFO, 1,
             _("     Recursion turned off. Will not descend from %s into %s\n"),
             ff_pkt->top_fname, ff_pkt->fname);
        ff_pkt->type = FT_DIREND; /* Backup only the directory entry */
        Dmsg1(130, "found FT_DIREND (no recurse): %s\n", ff_pkt->link);
        break;
      case FT_NOFSCHG:
        /* Suppress message for /dev filesystems */
        if (!IsInFileset(ff_pkt)) {
          Jmsg(jcr, M_INFO, 1,
               _("     %s is a different filesystem. Will not descend from %s "
                 "into it.\n"),
               ff_pkt->fname, ff_pkt->top_fname);
        }
        ff_pkt->type = FT_DIREND; /* Backup only the directory entry */
        Dmsg1(130, "found FT_DIREND (no fs change): %s\n", ff_pkt->link);
        break;
      case FT_INVALIDFS:
        Jmsg(
            jcr, M_INFO, 1,
            _("     Disallowed filesystem. Will not descend from %s into %s\n"),
            ff_pkt->top_fname, ff_pkt->fname);
        ff_pkt->type = FT_DIREND; /* Backup only the directory entry */
        Dmsg1(130, "found FT_DIREND (invalid fs): %s\n", ff_pkt->link);
        break;
      case FT_INVALIDDT:
        Jmsg(jcr, M_INFO, 1,
             _("     Disallowed drive type. Will not descend into %s\n"),
             ff_pkt->fname);
        break;
      case FT_REPARSE:
      case FT_JUNCTION:
      case FT_DIREND:
        Dmsg1(130, "found FT_DIREND: %s\n", ff_pkt->link);
        break;
      case FT_SPEC:
        Dmsg1(130, "found FT_SPEC: %s\n", ff_pkt->fname);
        if (S_ISSOCK(ff_pkt->statp.st_mode)) {
          Jmsg(jcr, M_SKIPPED, 1, _("     Socket file skipped: %s\n"),
               ff_pkt->fname);
          result.skip();
          return 1;
        }
        break;
      case FT_RAW:
        Dmsg1(130, "found FT_RAW: %s\n", ff_pkt->fname);
        break;
      case FT_FIFO:
        Dmsg1(130, "found FT_FIFO: %s\n", ff_pkt->fname);
        break;
      case FT_NOACCESS: {
        BErrNo be;
        Jmsg(jcr, M_NOTSAVED, 0, _("     Could not access \"%s\": ERR=%s\n"),
             ff_pkt->fname, be.bstrerror(ff_pkt->ff_errno));
        jcr->JobErrors++;
        result.skip();
        return 1;
      }
      case FT_NOFOLLOW: {
        BErrNo be;
        Jmsg(jcr, M_NOTSAVED, 0,
             _("     Could not follow link \"%s\": ERR=%s\n"), ff_pkt->fname,
             be.bstrerror(ff_pkt->ff_errno));
        jcr->JobErrors++;
        result.skip();
        return 1;
      }
      case FT_NOSTAT: {
        BErrNo be;
        Jmsg(jcr, M_NOTSAVED, 0, _("     Could not stat \"%s\": ERR=%s\n"),
             ff_pkt->fname, be.bstrerror(ff_pkt->ff_errno));
        jcr->JobErrors++;
        result.skip();
        return 1;
      }
      case FT_DIRNOCHG:
      case FT_NOCHG:
        Jmsg(jcr, M_SKIPPED, 1, _("     Unchanged file skipped: %s\n"),
             ff_pkt->fname);
        result.skip();
        return 1;
      case FT_ISARCH:
        Jmsg(jcr, M_NOTSAVED, 0, _("     Archive file not saved: %s\n"),
             ff_pkt->fname);
        return 1;
      case FT_NOOPEN: {
        BErrNo be;
        Jmsg(jcr, M_NOTSAVED, 0,
             _("     Could not open directory \"%s\": ERR=%s\n"), ff_pkt->fname,
             be.bstrerror(ff_pkt->ff_errno));
        jcr->JobErrors++;
        result.skip();
        return 1;
      }
      case FT_DELETED:
        Dmsg1(130, "FT_DELETED: %s\n", ff_pkt->fname);
        break;
      default:
        Jmsg(jcr, M_NOTSAVED, 0,
             _("     Unknown file type %d; not saved: %s\n"), ff_pkt->type,
             ff_pkt->fname);
        jcr->JobErrors++;
        result.skip();
        return 1;
    }

    try {
      // take note that ff_pkt->fname actually gets copied here since
      // stated_file uses a std::string instead of a char* to save the file
      // name!
      result.emplace_back(stated_file{
          ff_pkt->fname, ff_pkt->statp, ff_pkt->delta_seq, ff_pkt->type,
          ff_pkt->volhas_attrlist ? std::make_optional(ff_pkt->hfsinfo)
                                  : std::nullopt});
    } catch (...) {
      result.skip();
      return 0;
    }
    return 1;
  };
}

// todo: replace by synchronized
struct guarded_in_chan {
  guarded_in_chan(channel::in<found_files_list> in) : in{std::move(in)} {}
  std::mutex mut{};
  channel::in<found_files_list> in;
};

template <typename Deleter>
static void ListFromIncexe(JobControlRecord* jcr,
                           std::unique_ptr<FindFilesPacket, Deleter> ff_ptr,
                           std::shared_ptr<guarded_in_chan> in_ptr,
                           int fileset_idx)
{
  FindFilesPacket* ff = ff_ptr.get();
  findFILESET* fileset = ff->fileset;
  SetJcrInThreadSpecificData(jcr);
  findIncludeExcludeItem* incexe = fileset->include_list.get(fileset_idx);
  SetupLastOptionBlock(ff, incexe);
  // we do not need to follow hardlinks, as they will be handled
  // by the sending thread
  SetBit(FO_NO_HARDLINK, ff->flags);

  Dmsg4(50, "Verify=<%s> Accurate=<%s> BaseJob=<%s> flags=<%d>\n",
        ff->VerifyOpts, ff->AccurateOpts, ff->BaseJobOpts, ff->flags);

  dlistString* node;
  found_files_list result{static_cast<std::size_t>(fileset_idx)};
  foreach_dlist (node, &incexe->name_list) {
    char* fname = (char*)node->c_str();
    Dmsg1(debuglevel, "F %s\n", fname);
    ff->top_fname = fname;
    if (FindOneFile(jcr, ff, CreateCallback(SaveInList(result)), ff->top_fname,
                    (dev_t)-1, true)
        == 0) {
      result.error();
      break;
    }
    if (jcr->IsJobCanceled()) {
      result.error();
      break;
    }
  }

  result.sort();
  std::unique_lock lock(in_ptr->mut);
  in_ptr->in.put(std::move(result));
}

std::optional<list_files_threads> ListFiles(
    JobControlRecord* jcr,
    findFILESET* fileset,
    bool incremental,
    time_t save_time,
    std::optional<bool (*)(JobControlRecord*, FindFilesPacket*)> check_changed)
{
  if (fileset) {
    struct ff_cleanup {
      void operator()(FindFilesPacket* ff) { TermFindFiles(ff); }
    };

    auto [in, out] = channel::CreateBufferedChannel<found_files_list>(50);

    auto in_ptr = std::make_shared<guarded_in_chan>(std::move(in));
    list_files_threads threads{std::move(out)};

    for (int i = 0; i < fileset->include_list.size(); i++) {
      std::unique_ptr<FindFilesPacket, ff_cleanup> ff_pkt{init_find_files(),
                                                          ff_cleanup{}};
      ClearAllBits(FO_MAX, ff_pkt->flags);
      ff_pkt->fileset = fileset;
      SetFindOptions(ff_pkt.get(), incremental, save_time);
      if (check_changed)
        SetFindChangedFunction(ff_pkt.get(), check_changed.value());

      threads.emplace_thread(ListFromIncexe<ff_cleanup>, jcr, std::move(ff_pkt),
                             in_ptr, i);
    }

    return std::make_optional<list_files_threads>(std::move(threads));
  }
  return std::nullopt;
}

int SendPluginInfo(JobControlRecord* jcr,
                   FindFilesPacket* ff,
                   int PluginSave(JobControlRecord*, FindFilesPacket*, bool))
{
  findFILESET* fileset = ff->fileset;
  for (int i = 0; i < fileset->include_list.size(); i++) {
    dlistString* node;
    findIncludeExcludeItem* incexe
        = (findIncludeExcludeItem*)fileset->include_list.get(i);
    SetupLastOptionBlock(ff, incexe);
    foreach_dlist (node, &incexe->plugin_list) {
      char* fname = node->c_str();

      if (!PluginSave) {
        Jmsg(jcr, M_FATAL, 0, _("Plugin: \"%s\" not found.\n"), fname);
        return 0;
      }
      Dmsg1(debuglevel, "PluginCommand: %s\n", fname);
      ff->top_fname = fname;
      ff->cmd_plugin = true;
      PluginSave(jcr, ff, true);
      ff->cmd_plugin = false;
      if (jcr->IsJobCanceled()) { return 0; }
    }
  }

  return 1;
}

static bool CanBeHardLinked(struct stat& statp)
{
  bool hardlinked = false;
  switch (statp.st_mode & S_IFMT) {
    case S_IFREG:
    case S_IFCHR:
    case S_IFBLK:
    case S_IFIFO:
#ifdef S_IFSOCK
    case S_IFSOCK:
#endif
    {
      hardlinked = true;
    } break;
  }
  return hardlinked;
}

static bool SetupLink(FindFilesPacket* ff)
{
  switch (ff->statp.st_mode & S_IFMT) {
    case S_IFDIR: {
      std::size_t len = strlen(ff->fname);
      std::size_t max_len = len + 2;  // for trailing / and 0
      ff->link = (char*)malloc(max_len);
      if (!ff->link) return false;
      bstrncpy(ff->link, ff->fname, max_len);

      // strip all trailing slashes and then add one
      while (len >= 1 && IsPathSeparator(ff->link[len - 1])) { len--; }
      ff->link[len++] = '/'; /* add back one */
      ff->link[len] = '\0';
    } break;
    case S_IFLNK: {
      ff->link = (char*)malloc(path_max + name_max + 102);
      if (!ff->link) return false;
      int size = readlink(ff->fname, ff->link, path_max + name_max + 101);
      if (size < 0) {
        free(ff->link);
        return false;
      }
      ff->link[size] = 0;
    } break;
    default: {
      // this check is *should* be the same as ff->type != LNKSAVED
      if (ff->link == nullptr) ff->link = ff->fname;
    } break;
  }
  return true;
}

static void CleanupLink(FindFilesPacket* ff)
{
  switch (ff->statp.st_mode & S_IFMT) {
    case S_IFDIR:
    case S_IFLNK: {
      free(ff->link);
    } break;
  }
}

static bool SetupFFPkt(JobControlRecord* jcr,
                       FindFilesPacket* ff,
                       char* fname,
                       struct stat& statp,
                       int delta_seq,
                       int type,
                       std::optional<HfsPlusInfo> hfsinfo)
{
  ff->fname = fname;
  ff->statp = statp;
  ff->delta_seq = delta_seq;
  ff->type = type;

  ff->LinkFI = 0;
  ff->no_read = false;
  ff->linked = nullptr;
  ff->link = nullptr;  // we may need to allocate memory here instead
  if (!BitIsSet(FO_NO_HARDLINK, ff->flags) && statp.st_nlink > 1
      && CanBeHardLinked(statp)) {
    CurLink* hl = lookup_hardlink(jcr, ff, statp.st_ino, statp.st_dev);
    if (hl) {
      /* If we have already backed up the hard linked file don't do it
       * again */
      if (bstrcmp(hl->name, fname)) {
        Dmsg2(400, "== Name identical skip FI=%d file=%s\n", hl->FileIndex,
              fname);
        return false;
      } else {
        if (hl->FileIndex) {
          ff->link = hl->name;
          ff->type = FT_LNKSAVED; /* Handle link, file already saved */
          ff->LinkFI = hl->FileIndex;
          ff->linked = NULL;
          ff->digest = hl->digest;
          ff->digest_stream = hl->digest_stream;
          ff->digest_len = hl->digest_len;

          Dmsg3(400, "FT_LNKSAVED FI=%d LinkFI=%d file=%s\n", ff->FileIndex,
                hl->FileIndex, hl->name);
        } else {
          // if digest does not exist then whatever file created the
          // hardlink was not backed up (correctly). Try again here:

          if (ff->type == FT_LNKSAVED) {
            // this should only happen if something went wrong.
            // we cannot base our hardlink on FT_LNKSAVED
            // as that will not send any data.  We just have to
            // throw an error here
            return false;
          }

          int len = strlen(fname) + 1;
          hl->name = (char*)ff->linkhash->hash_malloc(len);
          bstrncpy(hl->name, fname, len);
          ff->linked = hl; /* Mark saved link */
          Dmsg2(400, "Added to hash FI=%d file=%s\n", ff->FileIndex, hl->name);
        }
      }
    } else {
      // File not previously dumped. Chain it into our list.
      hl = new_hardlink(jcr, ff, fname, ff->statp.st_ino, ff->statp.st_dev);
      ff->linked = hl; /* Mark saved link */
      Dmsg2(400, "Added to hash FI=%d file=%s\n", ff->FileIndex, hl->name);
    }
  }

  if (!SetupLink(ff)) return false;

  if (hfsinfo) {
    ff->volhas_attrlist = true;
    ff->hfsinfo = hfsinfo.value();
  } else {
    ff->volhas_attrlist = false;
  }

  return true;
}

struct stated_opened_file {
  stated_file f;
  std::optional<BareosFilePacket> bfd;
  int fileset;
};

void PrepareFileForSending(JobControlRecord* jcr,
                           std::vector<channel::out<stated_file>> outs,
                           channel::in<stated_opened_file> in)
{
  SetJcrInThreadSpecificData(jcr);
  std::size_t closed_channels = 0;
  while (closed_channels != outs.size()) {
    bool found_file = false;
    for (int i = 0; i < (int)outs.size(); ++i) {
      auto& out = outs[i];
      if (out.empty()) continue;
      std::optional<stated_file> file;
      while ((file = out.try_get())) {
        found_file = true;
        stated_file& f = file.value();
        BareosFilePacket bfd;
        binit(&bfd);
        // if (BitIsSet(FO_PORTABLE, f.flags)) {
        // 	SetPortableBackup(&bfd); /* disable Win32 BackupRead() */
        // }
        bool do_read = false;
        {
          /* Open any file with data that we intend to save, then save it.
           *
           * Note, if is_win32_backup, we must open the Directory so that
           * the BackupRead will save its permissions and ownership streams. */
          if (f.type != FT_LNKSAVED && S_ISREG(f.statp.st_mode)) {
#ifdef HAVE_WIN32
            do_read = !IsPortableBackup(&bfd) || f.statp.st_size > 0;
#else
            do_read = f.statp.st_size > 0;
#endif
          } else if (f.type == FT_RAW || f.type == FT_FIFO
                     || f.type == FT_REPARSE || f.type == FT_JUNCTION
                     || (!IsPortableBackup(&bfd) && f.type == FT_DIREND)) {
            do_read = true;
          }
        }
        // one can only safely read from a fifo
        // with a timer.  Otherwise we can get stuck here
        // as such we do not handle fifos here
        if (do_read && f.type != FT_FIFO) {
          // int noatime = BitIsSet(FO_NOATIME, f.flags) ? O_NOATIME : 0;
          int noatime = 0;
          if (bopen(&bfd, f.name.c_str(), O_RDONLY | O_BINARY | noatime, 0,
                    f.statp.st_rdev)
              < 0) {
            BErrNo be;
            Jmsg(jcr, M_NOTSAVED, 0, _("     Cannot open \"%s\": ERR=%s.\n"),
                 f.name.c_str(), be.bstrerror());
            jcr->JobErrors++;
          } else {
            if (!in.put({f, std::make_optional(bfd), i})) { return; }
          }
        } else {
          if (!in.put({f, std::nullopt, i})) { return; }
        }
      }
      if (out.empty()) { closed_channels += 1; }
    }
    // if we have not gotten any file then sleep for a bit instead
    // of spinning here
    if (!found_file) {
      std::this_thread::sleep_for(std::chrono::milliseconds(30));
    }
  }
}

int SendFiles(JobControlRecord* jcr,
              FindFilesPacket* ff,
              found_files_list list_result,
              int FileSave(JobControlRecord*, FindFilesPacket* ff_pkt, bool),
              int PluginSave(JobControlRecord*, FindFilesPacket* ff_pkt, bool))
{
  /* This is the new way */
  findFILESET* fileset = ff->fileset;
  int ret_val = 1;
  if (fileset) {
    /* TODO: We probably need be move the initialization in the fileset loop,
     * at this place flags options are "concatenated" accross Include {} blocks
     * (not only Options{} blocks inside a Include{}) */
    ClearAllBits(FO_MAX, ff->flags);
    SendPluginInfo(jcr, ff, PluginSave);
    ClearAllBits(FO_MAX, ff->flags);

    // some values are not setup by AcceptFile / SetupFFPkt but instead
    // get set once per include block inside SetupLastOptionBlock.
    // Not all of these are used during the send, so we "cache" them
    // here once and then always reuse them.
    struct cached_vals {
      int StripPath;
    };
    // everything is set to 0
    std::vector<cached_vals> cached_values(fileset->include_list.size());
    for (std::size_t i = 0; i < cached_values.size(); ++i) {
      SetupLastOptionBlock(ff, fileset->include_list.get(i));
      cached_values[i].StripPath = ff->StripPath;
    }
    if (list_result.has_error()) { return 0; }
    ff->incexe = fileset->include_list.get(list_result.fileset_idx());
    for (auto& file : list_result) {
      char* fname = file.name.data();
      if (!SetupFFPkt(jcr, ff, fname, file.statp, file.delta_seq, file.type,
                      file.hfsinfo)) {
        Dmsg1(debuglevel, "Error: Could not setup ffpkt for file '%s'\n",
              ff->fname);
        ret_val = 0;
        break;
      }
      if (!AcceptFile(ff)) {
        Dmsg1(debuglevel, "Did not accept file '%s'; skipping.\n", ff->fname);
        continue;
      }

      ff->bfd = BareosFilePacket{};
      binit(&ff->bfd);
      if (!FileSave(jcr, ff, false)) {
        if (IsBopen(&ff->bfd)) { bclose(&ff->bfd); }
        CleanupLink(ff);
        Dmsg1(debuglevel, "Error: Could not save file %s", ff->fname);
        ret_val = 0;
        break;
      } else {
        CleanupLink(ff);
        if (ff->linked) { ff->linked->FileIndex = ff->FileIndex; }
      }
      if (jcr->IsJobCanceled()) {
        ret_val = 0;
        break;
      }
    }
  }
  // this will close the opener regardless of whether
  // there are still files getting listed or not
  // since currently the opener will spin if it isnt fed fast enough
  return ret_val;
}
